/*
	File:		MoreOSA.cpp

	Contains:	C++ object for running OSA scripts

	Author:		Jon Pugh <mailto:jonpugh@frostbitefalls.com>

	Revisions:	7/19/99 First release
	
	Usage:		This reusable C++ object can be used pretty much as is.  It can
				load scripts from resources or memory, text or compiled.  It compiles,
				it loads, it stores, it runs, it returns results.  What more 
				do you want?
				
				This isn't heavily commented, but it was holding up the release,
				and since I'm doing this for free, the comments are sparse.

 */

#include "MoreOSA.h"
#include <ASDebugging.h>
#include <ASRegistry.h>
#include <OSAGeneric.h>
#include <string.h>
#include <Memory.h>
#include <Resources.h>


//	Error handling code
//
//	This AnyError class has only a constructor and these Macros.
//	It should be replacable with a minimum of effort.

AnyError::AnyError(OSAError e, char *m)
{
	err = e;
	msg = m;
}	// AnyError

// local Zone handling stack based object

class Zoner {
	private:
	Zone	*fZone;
	
	public:
	Zoner(Zone *zone);
	~Zoner();
};


Zoner::Zoner(Zone *zone)
{
	fZone = GetZone();
	SetZone(zone);
}	// Zoner

Zoner::~Zoner()
{
	SetZone(fZone);
}	// ~Zoner


// local resource file handling stack based object

class Rezer {
	private:
	short	refnum;
	short	oldrefnum;
	
	public:
	Rezer(FSSpec *file, short permissions = fsRdPerm);
	~Rezer();
};


Rezer::Rezer(FSSpec *file, short permissions)
{
	oldrefnum = CurResFile();
	refnum = FSpOpenResFile(file, permissions);
	if (refnum == -1)
		THROW_IF_ERR(resNotFound);
}	// Rezer

Rezer::~Rezer()
{
	CloseResFile(refnum);
	UseResFile(oldrefnum);
}	// ~Rezer


// initialize all the data members to 0
ScriptObject::ScriptObject()
{
	fComponent = nil;
	fScriptID = kOSANullScript;
	fResultID = kOSANullScript;
	fFile.vRefNum = 0;
	fFile.parID = 0;
	fFile.name[0] = 0;
	fIdleProc = nil;
	fPersistentProperties = true;
	fZone = GetZone();
}	// ScriptObject


// dispose all the memory hogging data
ScriptObject::~ScriptObject()
{
	OSADispose(GetComponent(), fScriptID);
	fScriptID = kOSANullScript;
	OSADispose(GetComponent(), fResultID);
	fResultID = kOSANullScript;
	CloseComponent();
}	// ~ScriptObject


// the component will be opened automatically when needed		
void ScriptObject::OpenComponent()
{
	if (fComponent == nil) {
		Zoner	z(fZone);
		fComponent = OpenDefaultComponent(kOSAComponentType, kOSAGenericScriptingComponentSubtype);
		if (fComponent == (ComponentInstance)badComponentInstance || fComponent == (ComponentInstance)badComponentSelector)
			fComponent = nil;
	}
}	// OpenComponent


void ScriptObject::CloseComponent()
{
	::CloseComponent(fComponent);
}	// CloseComponent


ComponentInstance ScriptObject::GetComponent()
{
	OpenComponent();
	return fComponent;
}	// GetComponent


void ScriptObject::UseZone(Zone *zone)
{
	fZone = zone;
}	// UseZone


void ScriptObject::UseSystemZone()
{
	fZone = SystemZone();
}	// UseSystemZone


// use Compile or Load to initialize a script
Boolean ScriptObject::Compile(FSSpec *file)
{
	long	size;
	Handle	data;
	short	refnum;
	Boolean	success;
	Zoner	z(fZone);
		
	fFile = *file;
	THROW_IF_ERR( FSpOpenDF(file, fsRdPerm, &refnum) );
	TRY {
		THROW_IF_ERR( GetEOF(refnum, &size) );
		data = NewHandle(size);
		THROW_IF_NIL(data);
		HLock(data);
		THROW_IF_ERR( FSRead(refnum, &size, *data) );
		HUnlock(data);
		THROW_IF_ERR( FSClose(refnum) );
		success = Compile(typeChar, data);
		DisposeHandle(data);
	}
	CATCH {
		FSClose(refnum);
		DisposeHandle(data);
		RERAISE;
	}
	return success;
}	// Compile


Boolean ScriptObject::Compile(DescType type, Handle scriptText)
{
	AEDesc	script;
	Zoner	z(fZone);
		
	script.dataHandle = scriptText;
	script.descriptorType = type;
	return OSACompile(GetComponent(), &script, kOSAModeCompileIntoContext, &fScriptID) == noErr;
}	// Compile



// save the formatted script text
void ScriptObject::StoreSource()
{
	StoreSource(&fFile);
}	// StoreSource


void ScriptObject::StoreSource(FSSpec *file)
{
	AEDesc	text;
	long	size;
	short	refnum;
	
	THROW_IF_ERR( FSpOpenDF(file, fsRdWrPerm, &refnum) );
	GetSource(typeChar, &text);
	TRY {
		THROW_IF_ERR( SetEOF(refnum, 0) );
		HLock(text.dataHandle);
		size = GetHandleSize(text.dataHandle);
		THROW_IF_ERR( FSWrite(refnum, &size, *text.dataHandle) );
		//HUnlock(text.dataHandle);
		THROW_IF_ERR( SetEOF(refnum, size) );
		THROW_IF_ERR( FSClose(refnum) );
		AEDisposeDesc(&text);
	}
	CATCH {
		FSClose(refnum);
		AEDisposeDesc(&text);
	}
}	// StoreSource



// load a compiled script into memory
void ScriptObject::Load(AliasHandle alias)
{
	FSSpec	file;
	Boolean	wasChanged;
	
	THROW_IF_ERR( ResolveAlias(nil, alias, &file, &wasChanged) );
	Load(&file);
}	// Load


// load a compiled script into memory
void ScriptObject::Load(FSSpec *file)
{
	Handle	script;
	Zoner	z(fZone);
	Rezer	rz(file);
	
	fFile = *file;
	script = GetResource(kOSAScriptResourceType, 128);
	if (script == nil)
		script = Get1IndResource(kOSAScriptResourceType, 1);
	if (script == nil)
		script = GetIndResource(kOSAScriptResourceType, 1);
	if (script == nil) {
		THROW_IF_ERR(resNotFound);
	}
	Load(script);
}	// Load


void ScriptObject::Load(Handle scriptData)
{
	AEDesc	script;
	Zoner	z(fZone);
	
	script.dataHandle = scriptData;
	script.descriptorType = typeOSAGenericStorage;
	THROW_IF_ERR( OSALoad(GetComponent(), &script, kOSANullMode, &fScriptID) );
}	// Load


// has a script been loaded?
Boolean ScriptObject::ScriptLoaded()
{
	return fScriptID != kOSANullScript;
}	// ScriptLoaded


// set a number of optional callbacks
// idleProc should be used or events will be lost
// a better solution is to provide your own sendProc & activeProc instead
void ScriptObject::SetIdleProc(AEIdleUPP idleProc)
{
	fIdleProc = idleProc;
}	// SetIdleProc


// send and active procs provide time while scripts run
void ScriptObject::SetSendProc(OSASendUPP sendProc, long refcon)
{
	// error if fIdleProc != nil?  idleProc will be ignored
	THROW_IF_ERR( OSASetSendProc(GetComponent(), sendProc, refcon) );
}	// SetSendProc


void ScriptObject::SetActiveProc(OSAActiveUPP activeProc, long refcon)
{
	// error if fIdleProc != nil?  idleProc will be ignored
	THROW_IF_ERR( OSASetActiveProc(GetComponent(), activeProc, refcon) );
}	// SetActiveProc


// these provide additional less used hooks
void ScriptObject::SetResumeProc(AEEventHandlerUPP resumeProc, long refcon)
{
	THROW_IF_ERR( OSASetResumeDispatchProc(GetComponent(), resumeProc, refcon) );
}	// SetResumeProc


void ScriptObject::SetCreateProc(OSACreateAppleEventUPP createProc, long refcon)
{
	THROW_IF_ERR( OSASetCreateProc(GetComponent(), createProc, refcon) );
}	// SetCreateProc



// control whether Run methods call Store automatically
Boolean ScriptObject::GetPersistentProperties()
{
	return fPersistentProperties;
}	// GetPersistentProperties


void ScriptObject::SetPersistentProperties(Boolean persist)
{
	fPersistentProperties = persist;
}	// SetPersistentProperties



// run a script by executing the run handler
void ScriptObject::Run()
{
	Zoner	z(fZone);

	THROW_IF_ERR( OSAExecute(GetComponent(), fScriptID, kOSANullScript, kOSANullMode, &fResultID) );
}	// Run


void ScriptObject::Run(FSSpec *file)
{
	Load(file);
	Run();
	if (fPersistentProperties && IsModified())
		Store();
}	// Run


void ScriptObject::Run(Handle scriptData)
{
	Load(scriptData);
	Run();
	if (fPersistentProperties && IsModified())
		Store(scriptData);
}	// Run


Boolean ScriptObject::IsModified()
{
	long	result;
	Zoner	z(fZone);	// ???
	
	THROW_IF_ERR( OSAGetScriptInfo(GetComponent(), fScriptID, kOSAScriptIsModified, &result) );
	return result > 0;
}	// IsModified


// list all handlers within a script
// all names are in lowercase
void ScriptObject::GetHandlerNames(AEDescList* handlers)
{
	ComponentInstance	as;
	OSAID				asID;
	Zoner				z(fZone);
	
	asID = fScriptID;
	if ( ScriptLoaded() ) {
		THROW_IF_ERR( OSAGenericToRealID(GetComponent(), &asID, &as) );
		THROW_IF_ERR( OSAGetHandlerNames(as, kOSANullMode, asID, handlers) );
	}
	else {
		// no script, make empty list
		THROW_IF_ERR( AECreateList(NULL, 0, false, handlers) );
	}
}	// GetHandlerNames


// preflight if a handler is present in a script
// the input name should be lowercased for AppleScript english scripts
Boolean ScriptObject::HasHandler(char *handlerName)
{
	AEDesc		handlers;
	long		n, i, size;
	DescType	type, key;
	char		handler[256];
	Zoner		z(fZone);
	Boolean		hasHandler;
	OSErr		err;
	
	GetHandlerNames(&handlers);
	TRY {
		THROW_IF_ERR( AECountItems(&handlers, &n) );
		hasHandler = false;
		for (i = 1; i <= n; i++) {
			err = AEGetNthPtr(&handlers, i, typeChar, &key, &type, 
					&handler, sizeof(handler), &size);
			if (err == noErr && strncmp(handlerName, (char*)&handler, size) == 0) {
				hasHandler = true;
				break;
			}
		}
		AEDisposeDesc(&handlers);
	}
	CATCH {
		AEDisposeDesc(&handlers);
	}
	return hasHandler;
}	// HasHandler

// the handler name should be lowercased in AppleScript English
// the params parameters should be a list of positional parameters
// there is no provision for named parameters in this release
void ScriptObject::CallSubroutine(char *handlerName, AEDescList *params)
{
	AppleEvent				event;
	AEDesc					address;
	ProcessSerialNumber		psn;
	Zoner					z(fZone);
	
	psn.highLongOfPSN = 0;
	psn.lowLongOfPSN = kCurrentProcess;
	THROW_IF_ERR( AECreateDesc(typeProcessSerialNumber, &psn, sizeof(psn), &address) );
	THROW_IF_ERR( AECreateAppleEvent(kASAppleScriptSuite, kASSubroutineEvent, 
			&address, kAutoGenerateReturnID, kAnyTransactionID, &event) );
	THROW_IF_ERR( AEDisposeDesc(&address) );
	THROW_IF_ERR( AEPutParamPtr(&event, keyASSubroutineName, typeChar, handlerName, strlen(handlerName)) );
	if (params != nil)
		THROW_IF_ERR( AEPutParamDesc(&event, keyDirectObject, params) );
	DoEvent(&event);
	AEDisposeDesc(&event);
	if (fPersistentProperties && IsModified())
		Store();
}	// CallSubroutine


const int kExecuteMode = kOSAModeCanInteract | kAECanSwitchLayer;


// send an event to the script
void ScriptObject::DoEvent(AppleEvent *event)
{
	Zoner	z(fZone);

	THROW_IF_ERR(OSAExecuteEvent(GetComponent(), event, fScriptID, kExecuteMode, &fResultID));
}	// DoEvent


void ScriptObject::DoEvent(AppleEvent *event, AppleEvent *reply)
{
	Zoner	z(fZone);

	THROW_IF_ERR( OSADoEvent(GetComponent(), event, fScriptID, kExecuteMode, reply) );
}	// DoEvent



// access info about the script
void ScriptObject::GetSource(AEDesc *text)
{
	GetSource(typeChar, text);
}	// GetSource


void ScriptObject::GetSource(DescType desiredType, AEDesc *text)
{
	Zoner	z(fZone);

	THROW_IF_ERR( OSAGetSource(GetComponent(), fScriptID, desiredType, text) );
}	// GetSource


void ScriptObject::ReturnResults(AEDesc *result)
{
	ReturnResults(typeWildCard, result);
}	// ReturnResults


void ScriptObject::ReturnResults(DescType desiredType, AEDesc *result)
{
	Zoner	z(fZone);

	THROW_IF_ERR( OSACoerceToDesc(GetComponent(), fResultID, desiredType, kOSANullMode, result) );
}	// ReturnResults


void ScriptObject::ReturnResultText(AEDesc *result)
{
	Zoner	z(fZone);

	THROW_IF_ERR( OSADisplay(GetComponent(), fResultID, typeChar, kOSAModeDisplayForHumans, result) );
}	// ReturnResults


void ScriptObject::GetErrorMessage(DescType desiredType, AEDesc *result)
{
	Zoner	z(fZone);

	THROW_IF_ERR( OSAScriptError(GetComponent(), kOSAErrorMessage, desiredType, result) );
}	// GetError


long ScriptObject::GetErrorNumber()
{
	AEDesc	result;
	long	temp;
	Zoner	z(fZone);
	
	THROW_IF_ERR( OSAScriptError(GetComponent(), kOSAErrorNumber, typeInteger, &result) );
//	ASSERT(GetHandleSize(result.dataHandle) == sizeof(long));
	temp = *(long*)*result.dataHandle;
	AEDisposeDesc(&result);
	return temp;
}	// GetError


void ScriptObject::GetErrorRange(AEDesc *result)
{
	Zoner	z(fZone);

	THROW_IF_ERR( OSAScriptError(GetComponent(), kOSAErrorRange, typeOSAErrorRange, result) );
}	// GetErrorRange



// save the original
void ScriptObject::Store()
{
	Store(&fFile);
}	// Store


void ScriptObject::Store(FSSpec *file)
{
	Handle	script;
	Zoner	z(fZone);
	Rezer	rz(file, fsRdWrPerm);
	
	script = GetResource(kOSAScriptResourceType, 128);
	if (script == nil)
		script = Get1IndResource(kOSAScriptResourceType, 1);
	if (script == nil)
		script = GetIndResource(kOSAScriptResourceType, 1);
	if (script == nil) {
		THROW_IF_ERR(resNotFound);
	}
	Store(script);
	ChangedResource(script);
}	// Store


Handle ScriptObject::Store(Handle oldScript)
{
	AEDesc	desc;
	long	size;
	Zoner	z(fZone);
	
	THROW_IF_ERR( OSAStore(GetComponent(), fScriptID, typeOSAGenericStorage, kOSAModeNull, &desc) );
	if (oldScript == nil)
		return desc.dataHandle;
	else {
		size = GetHandleSize(desc.dataHandle);
		SetHandleSize(oldScript, size);
		BlockMoveData(*desc.dataHandle, *oldScript, size);
		return oldScript;
	}
}	// Store


// save the original if it changed
void ScriptObject::StoreIfChanged()
{
	if (fPersistentProperties && IsModified())
		Store(&fFile);
}	// Store


void ScriptObject::StoreIfChanged(FSSpec *file)
{
	if (fPersistentProperties && IsModified())
		Store(file);
}	// Store


Handle ScriptObject::StoreIfChanged(Handle oldScript)
{
	if (fPersistentProperties && IsModified())
		oldScript = Store(oldScript);
	return oldScript;
}	// Store


