/*
 * Block Read.cp
 */

#include "FileLib.h"
#include "Log.h"
#include "FSM.h"
#include "ISO9660tt.h"
#include "ISO Reader.h"
#include "Block Read.h"

#define RoundDown(val,mark) (val - val % mark)
#define RoundUp(val,mark) ((val % mark)? (RoundDown(val,mark) + mark) : val)


#ifdef READ_FROM_FILE
	static	short	gFileRefNum = 0;
#endif

OSErr DeviceRead (short drive, short refnum, long start, long bytes, Ptr buffer, short cacheOptions)
{
	assert (start % 512 == 0);
	assert (bytes % 512 == 0);
	ParamBlockRec	params;
#ifdef READ_FROM_FILE
	refnum = gFileRefNum; drive = 0;
#endif
	params.ioParam.ioCompletion = 0;
	params.ioParam.ioVRefNum = drive;
	params.ioParam.ioRefNum = refnum;
	params.ioParam.ioBuffer = buffer;
	params.ioParam.ioReqCount = bytes;
	params.ioParam.ioPosMode = fsFromStart | cacheOptions;
	params.ioParam.ioPosOffset = start;
	return PBReadSync (&params);
}

/*
 * Out Little Cache:
 */
#define	BLKS	2
typedef char	SectorBuffer[ISO_LOGICAL_SECTOR_SIZE];
static	short			blkInUse[BLKS] = {0};
static	SectorBuffer	gSectorBuf[BLKS];
static	long			bufferedSecNum[BLKS];
static	short			bufferedVolID[BLKS] = {0};


static Boolean FindSectorBuffer (long sectorNumber, Ptr &p)
// Looks if the desired sector is in our little cache.
// If so, this sector will be re-used, otherwise it returns false
{
	for (int i = 0; i < BLKS; ++i) {
		if (bufferedSecNum[i] == sectorNumber && bufferedVolID[i] == isoGlobals->vRefNum) {
			blkInUse[i] += 1;
			p = gSectorBuf[i];
			return true;
		}
	}
	return false;
}

Boolean GetSectorBuffer (Ptr &p)
// allocates a buffer for a new sector that is not in our cache yet.
{
	for (int i = 0; i < BLKS; ++i) {
		if (!blkInUse[i]) {
			blkInUse[i] = 1;
			p = gSectorBuf[i];
			bufferedVolID[i] = 0;
			bufferedSecNum[i] = -1;
			return true;
		}
	}
	Fatal ("No free sector buf");
	return false;
}

static void ValidateSectorBuffer (Ptr &p, long sectorNumber)
{
	for (int i = 0; i < BLKS; ++i) {
		if (p == gSectorBuf[i]) {
			bufferedVolID[i] = isoGlobals->vRefNum;
			bufferedSecNum[i] = sectorNumber;
			return;
		}
	}
	Fatal ("No sector to mark");
}

void ReleaseSectorBuffer (Ptr p)
{
	for (int i = 0; i < BLKS; ++i) {
		if (p == gSectorBuf[i]) {
			assert (blkInUse[i] > 0);
			--blkInUse[i];
			return;
		}
	}
	Fatal ("No sector buf to release");
}

void FlushSectorBuffers (void)
{
	for (int i = 0; i < BLKS; ++i) {
		bufferedSecNum[i] = -1;
		bufferedVolID[i] = 0;
	}
}

short FreeSectorBuffers ()
{
	int n = 0;
	for (int i = 0; i < BLKS; ++i) {
		if (!blkInUse[i]) {
			++n;
		}
	}
	return n;
}

extern VCBPtr CurrVCB;
extern unsigned long fsmVersion;

static OSErr ReadPhysSector (long sectorNumber, Ptr sectorBuffer, Boolean temporary)
// alloziert keinen Block sondern liest ihn nach 'blockBuf' ein.
{
	OSErr err;
	#ifdef READ_FROM_FILE
		#pragma unused (temporary)
		err = DeviceRead (0, gFileRefNum, sectorNumber * isoGlobals->sectorSize, isoGlobals->sectorSize, sectorBuffer, 0);
	#else
		long blks = (long)isoGlobals->sectorSize >> 9;
		long blkNum = sectorNumber * blks;
		if (fsmVersion >= 0x200) {
			// use the new volume-level multiple-block read function
			// would also work with FSMVersion >= 0x130, but would crash with Speed Doubler 8.0
			unsigned long n;
			err = UTVolCacheReadIP (CurrVCB, blkNum, sectorBuffer, isoGlobals->sectorSize, &n, (short)(temporary?noCacheMask:0));
			//LogPStr (err?LKError:LKUser1, nil, "\pVolCacheReadIP blk ^0, req ^1, act ^2, err ^3", (long)blkNum/blks, (long)vol.blockSize>>9, (long)n>>9, (long)err);
			if (err > 0) err = ioErr;
			if (n < isoGlobals->sectorSize && !err) err = ioErr;
		} else {
			// use the volume-level single-block read function
			long pos = 0;
			do {
				Ptr ptr;
				err = UTGetBlock (isoGlobals->vRefNum, NULL, blkNum, gbDefault, &ptr);
				if (err > 0) err = ioErr;
				if (err) break;
				BlockMoveData (ptr, &sectorBuffer[pos], 512);
				UTReleaseBlock (ptr, (temporary)?rbFreeMask:rbDefault);
				++blkNum;
				pos += 512;
			} while (--blks);
		}
	#endif
	AssertNum (!err, err);
	return err;
}


OSErr GetDirSector (long sectorNumber, void* &sectorAddress, Boolean temporary)
{
	OSErr err;
	Ptr sectorBuffer;
	if (FindSectorBuffer (sectorNumber, sectorBuffer)) {
		// we got this sector still in our little cache
		err = noErr;
	} else {
		if (!GetSectorBuffer (sectorBuffer)) {
			return -1;	// internal error
		}
		err = ReadPhysSector (sectorNumber, sectorBuffer, temporary);
		if (err) {
			ErrorNum ("reading dir sector failed", sectorNumber);
			ReleaseSectorBuffer (sectorBuffer);
		} else {
			ValidateSectorBuffer (sectorBuffer, sectorNumber);
		}
	}
	if (!err) {
		sectorAddress = sectorBuffer;
	} else {
		sectorAddress = (Ptr)0xB5555577;
	}
	return err;
}

OSErr ReleaseDirSector (void* sectorAddress)
{
	ReleaseSectorBuffer ((Ptr)sectorAddress);
	return noErr;
}

/*
OSErr ReleaseLogSector (void* sectorAddress, SectorTypes secType)
{
	ReleaseSectorBuffer ((Ptr)sectorAddress);
	return noErr;
}
*/

void InitBlockRead ()
{
	#ifdef READ_FROM_FILE
		FSSpec spec;
		FSMakeFSSpec (0, 0, "\pTest-Image", &spec);
		FSpResolveAlias (spec);
		assert (FSpOpenDF (&spec, fsCurPerm, &gFileRefNum) == noErr);
		long pos; GetEOF (gFileRefNum, &pos);
		isoGlobals->maxBlockNumber = pos / ISO_LOGICAL_SECTOR_SIZE;
	#endif
}


// EOF
