/*
 * ISO Dir Locator.cp
 *
 * Directory Sector Locator Tables
 */

/*
Prinzipieller Aufbau der Tables:
Damit Speicherverwaltung einigermaen effektiv (schnell) ist, sollte die Gre
des bentigten Platzes pro Dir von vornherein festgelegt werden. Problem: Wieviel
Platz sollte fr den Dateinamen reserviert werden?
Mgliche Lsungen:
1. Fr den Namen werden immer 32 Byte belegt -> verschwenderisch
   Zusatzinfo: Ein ISO-Name kann auch lnger als 32 Byte sein: a) die Vers-Kennung wird
   noch hinzugezhlt, b) ISO-Erweiterungsformate lassen mehr Zeichen zu.
2. Es werden nur die ersten n Bytes jeden Namens belegt -> Kann ineffizient werden,
   Beispiel: Es werden nur 8 Byte gespeicht und alle Dateien beginnen mit "BILDARCHV"
   und dann folgt eine laufende Nummer.
3. Zur Umgehung des 2. Problems wird im 1. Byte des abgelegten Namens vermerkt,
   ab welcher Stelle es sich vom Vorgnger unterscheidet.
   So knnte schon bei 1+3 Byte eine gute Unterscheidung erreicht werden und
   es ist meistens hchstens ein einziger Vorgnger-Sektor wiederholt lesen,
   im Fall, da der Name nicht den Sektor eindeutig abgrenzt.
Alle Dir-Arrays werden hintereinander angelegt und bei Platzmangel wird einfach
das lteste Array entfernt und der Rest dahinter verschoben. Neue Arrays werden
damit immer am Ende der Arrays angefgt.

Aktuell implementierte Lsung: Nr. 2: es werden 13 Zeichen gespeichert.
*/

#include <stdlib.h>
#include <string.h>
#include "PascalLib.h"
#include "ISO Dir Locator.h"

typedef unsigned long ulong;
typedef unsigned short word;


/*
 * exported vars
 */
Boolean LocBufOverflow = false;


#pragma options align = mac68k
	
	typedef struct {
		long	firstIndex;			// index excl associated, incl self & par entries
		LocName	firstName;
	} LocEntry;
	
	typedef struct LocArray {
		LocArray	*next;			// mu erstes Feld sein, um Ende des belegten Puffers zu markieren
		long		appleDirID;		// is zero if unknown
		BlkNum		dirStartBlkNum;
		word		dirBlocks;		// 'dirBlk' array contains 'dirBlocks-1' elements
		short		volID;
		long		dirRecCount;	// no of entries excl associated, incl self & par entries
		ulong		timeStamp;		// increases each time this dir is accessed
		LocEntry	dirBlk[];		// '.startIndex==0' -> array not filled up yet
	} LocArray;

#pragma options align = reset


#define LALEN(dirLoc) (((LocArray*)dirLoc)->dirBlocks * sizeof (LocEntry) + (sizeof (LocArray) - sizeof (LocEntry)))


static Ptr	LocBufStart, LocBufEnd = nil;
static long LocBufSize = 0, LocBufUsed;

static ulong	TimeStampHi = 0;

static LocArray *findLowestTimeStamp (void)
{
	LocArray *s = (LocArray*)LocBufStart;
	LocArray *lp = s;
	ulong low = TimeStampHi;
	if (LocBufUsed) do {
		ulong ts = s->timeStamp;
		if (TimeStampHi - ts > TimeStampHi - low) {	// welcher Abstand zu TimeStampHi ist grer?
			low = ts;
			lp = s;
		}
		s = s->next;
	} while (s);
	return lp;
}

static LocArray *LocArrayLast = nil;
//static LocArray *LocArrayRover = nil;	// irgendwann mal sinnvoller setzen!!!
#define LocArrayRover LocArrayLast

static LocArray *findLastLocArray ()
{
	LocArray *p1 = (LocArray*)LocBufStart, *p2;
	assert (p1);
	do {
		p2 = p1->next;
		if (p2 == nil) return p1;
		p1 = p2;
	} while (true);
}

static Boolean makeRoomFor (long requiredSize, LocArray* &dirLoc)
{
	if ((LocBufSize - LocBufUsed) < requiredSize) {
		LocBufOverflow = true;
		if (requiredSize > LocBufSize) {
			return false;	// can't make room - buffer too small
		}
		// make room by deleting the oldest recorded dirs
		do {
			LocArray *p;
			p = findLowestTimeStamp ();
			long diff = LALEN (p);
			Ptr end = (Ptr)p + diff;
			BlockMoveData (end, (Ptr)p, LocBufStart + LocBufUsed - end);
			if (LocArrayLast == p) LocArrayLast = nil;
			LocBufUsed -= diff;
			assert (LocBufUsed >= 0);
		} while ((LocBufSize - LocBufUsed) < requiredSize);
		((LocArray*)(LocBufStart+LocBufUsed))->next = nil;	// Ende-Markierung
	}
	dirLoc = (LocArray*)&LocBufStart[LocBufUsed];
	if (LocArrayLast == nil) LocArrayLast = findLastLocArray ();
	assert (!LocArrayLast->next);
	LocArrayLast->next = dirLoc;
	LocArrayLast = dirLoc;
	dirLoc->next = nil;
	LocBufUsed += requiredSize;
	assert (LocBufUsed <= LocBufSize);
	return true;
}

static LocArray* locateArrayByDirID (short volID, long dirID)
{
	if (LocBufSize > 0) {
		LocArray	*dirLoc = (LocArray*) LocBufStart;
		if (IsAppleDirID (dirID)) {
			// search for appleDirID
			if (LocArrayRover && LocArrayRover->volID == volID && LocArrayRover->appleDirID == dirID) {
				return LocArrayRover;
			}
			if (LocBufUsed) do {
				if (dirLoc->volID == volID && dirLoc->appleDirID == dirID) {
					return dirLoc;
				}
			} while ((dirLoc = dirLoc->next) != nil);
		} else {
			// search for dirStartBlkNum
			BlkNum startBlk; long ofs;
			ResolveDirID (dirID, startBlk, ofs);
			assert (ofs < 0);
			if (LocArrayRover && LocArrayRover->dirStartBlkNum == startBlk && LocArrayRover->volID == volID) {
				return LocArrayRover;
			}
			if (LocBufUsed) do {
				if (dirLoc->dirStartBlkNum == startBlk && dirLoc->volID == volID) {
					return dirLoc;
				}
			} while ((dirLoc = dirLoc->next) != nil);
		}
	}
	return nil;
}

Boolean LocateBlkNumByName (short volID, long dirID, StringPtr nameToFind, DirLocRec &loc)
// The passed "nameToFind" must already be filtered with FilterMacName
{
	LocArray *dirLoc = locateArrayByDirID (volID, dirID);
	InitDirLocRec (loc);
	if (dirLoc) {
		long blkCnt = loc.dirTotalBlks = dirLoc->dirBlocks;
		long blkNumToStartWith = loc.dirStartBlk = dirLoc->dirStartBlkNum;
		long indexToStartWith = 0;
		long n = 0;
		if (isoGlobals->namesAreInOrder) {
			while (--blkCnt > 0) {
				LocName firstNameOfBlk;
				assert (sizeof(dirLoc->dirBlk[0].firstName) == sizeof(LocName));	// firstNameOfBlk must fit for firstName
				pstrcpy (firstNameOfBlk, dirLoc->dirBlk[n].firstName);
				if (firstNameOfBlk[0] == 0) {
					// end of known values - there can't be any further known indices
					break;
				} else {
					//### FilterISOName (firstNameOfBlk);
					if (CompareISONames (nameToFind, firstNameOfBlk, sizeof(LocName)-1) > 0) {
						// the name to find won't be before this block
						indexToStartWith = dirLoc->dirBlk[n].firstIndex;
						blkNumToStartWith = loc.dirStartBlk + (++n);
					} else {
						// we're over the mark - take values of previous iteration
						break;
					}
				}
			}
		}
		loc.currBlkNum = blkNumToStartWith;
		loc.currStartIdx = indexToStartWith;
		loc.helper = (long)dirLoc;
		return true;
	} else {
		return false;	// dir is not recorded at all -> must start to load dir from scratch
	}
}

Boolean LocateBlkNumByIndex (short volID, long dirID, long index, DirLocRec &loc)
{
	LocArray *dirLoc = locateArrayByDirID (volID, dirID);
	InitDirLocRec (loc);
	if (dirLoc) {
		long blkCnt = loc.dirTotalBlks = dirLoc->dirBlocks;
		long lastBlkNum = loc.dirStartBlk = dirLoc->dirStartBlkNum;
		long lastStartIdx = 0;
		long n = 0;
		while (--blkCnt > 0) {
			long startIndex = dirLoc->dirBlk[n++].firstIndex;
			if (startIndex >= index) {	// must be ">=", not ">", or Files wird Rsrc-Entry in Blk1, Data-Entry in Blk2 will not be found correctly!
				// we're over the mark - take previous values
				break;
			} else if (startIndex == 0) {
				// end of known values - there can't be any further known indices
				break;
			} else {
				lastStartIdx = startIndex;
				lastBlkNum = loc.dirStartBlk + n;
			}
		}
		loc.currBlkNum = lastBlkNum;
		loc.currStartIdx = lastStartIdx;
		loc.helper = (long)dirLoc;
		return true;
	} else {
		return false;	// dir is not recorded at all -> must start to load dir from scratch
	}
}

Boolean DirLocGetNoOfDirRecs (short volID, long dirID, long& count)
{
	LocArray *dirLoc = locateArrayByDirID (volID, dirID);
	if (dirLoc && dirLoc->dirRecCount) {
		count = dirLoc->dirRecCount;
		return true;
	} else {
		count = 0;
		return false;
	}
}

void InitDirLocRec (DirLocRec &loc)
{
	loc.dirStartBlk = 0;
	loc.helper = nil;
	loc.prevBlk = 0;
}

static void checkNameOrdering (LocName *prevName, Byte *dirSector)
{
	long offs = 0;
	DirRec* dirRecPtr = (DirRec*)dirSector;
	do {
		short res = CompareISONames ((StringPtr)prevName, &dirRecPtr->nameLen, sizeof(LocName));
		if (res > 0) {
			DebugStr ("\pWarning: Names are not in order!");
			isoGlobals->namesAreInOrder = false;
			break;
		}
		prevName = (LocName*)&dirRecPtr->nameLen;
	} while (AdvanceToNextDirRec (dirSector, offs, dirRecPtr));
}

static void getLastNameOfDirBlock (LocName *lastName, Byte *dirSector)
{
	long offs = 0;
	DirRec* dirRecPtr = (DirRec*)dirSector, *prevRecPtr;
	do {
		prevRecPtr = dirRecPtr;
	} while (AdvanceToNextDirRec (dirSector, offs, dirRecPtr));
	pstrncpy ((StringPtr)lastName, &prevRecPtr->nameLen, sizeof(*lastName));
}

void UpdateDirLocationBuffer (short volID, long dirID, Byte *dirSector, long entriesInBlk, DirLocRec& loc)
{
	/*
	 * 1. haben wir den Blk bereits eingetragen?
	 * 2. wenn nein: pat er in den puffer? falls nein: ignorieren und ende.
	 * 3. dirRecCount setzen, wenn letzter Block des Dirs gelesen wurde
	 * 4. Prfen, ob Reihenfolge aller eingetragenen Dateinamen stimmt, um ggf. die
	 *    indizierte Namenssuche ber LocateBlkNumByName zu deaktivieren (durch
	 *    namesAreInOrder=false). Es reicht, da diese Erkennnung erst hier und nur
	 *    hier gemacht wird, denn bevor es vorkommt, da ein Name ber LocateBlkNumByName
	 *    berhaupt gefunden werden kann, mu sein Dir-Block hier bekannt geworden sein.
	 */

	if (loc.dirStartBlk == 0) {
		// In this case we've got a block from Apple's dirID where we don't know its dir start
		return;
	}
	
	LocArray *dirLoc;
	if (loc.helper) {
		dirLoc = (LocArray*) loc.helper;
	} else {
		dirLoc = locateArrayByDirID (volID, dirID);
	}
	if (!dirLoc) {
		long totalSize = sizeof (LocArray) + (loc.dirTotalBlks-1) * sizeof (LocEntry);
		if (!makeRoomFor (totalSize, dirLoc)) {
			return;		// LocBuffer is too small - do not buffer this directory
		}
		// add this new directory to DirLoc buffer
		dirLoc->volID = volID;
		dirLoc->dirStartBlkNum = loc.dirStartBlk;
		dirLoc->appleDirID = (dirID == fsRtDirID) ? fsRtDirID : 0;
		dirLoc->dirBlocks = loc.dirTotalBlks;
		dirLoc->dirRecCount = 0;
		for (int i = 1; i < loc.dirTotalBlks; ++i) {
			dirLoc->dirBlk[i-1].firstIndex = 0;
			dirLoc->dirBlk[i-1].firstName[0] = '\0';
		}
	}
	// update this block
	assert (dirLoc->dirStartBlkNum = loc.dirStartBlk);
	assert (dirLoc->dirBlocks = loc.dirTotalBlks);
	dirLoc->timeStamp = TimeStampHi++;
	long blkDiff = loc.currBlkNum-loc.dirStartBlk;
	assert (blkDiff < loc.dirTotalBlks);
	// Namen & Index eintragen fr alle Blcke nach dem ersten Block eines Dirs
	if (blkDiff > 0) {
		if (dirLoc->dirBlk[blkDiff-1].firstIndex == 0) {
			// dieser Block ist noch nicht eintragen worden
			StringPtr firstName = &((DirRec*)dirSector)->nameLen;
			dirLoc->dirBlk[blkDiff-1].firstIndex = loc.currStartIdx;
			pstrncpy (dirLoc->dirBlk[blkDiff-1].firstName, firstName, sizeof(dirLoc->dirBlk[0].firstName));
			assert (loc.prevBlk == loc.currBlkNum-1);	// sonst kann die Namensordnung nicht berprft werden.
			checkNameOrdering (&loc.lastNameInPrevBlk, dirSector);
		} else {
			assert (dirLoc->dirBlk[blkDiff-1].firstIndex == loc.currStartIdx);
		}
	} else {
		assert (loc.currStartIdx == 0);
	}

	// letzten Namen des Dir-Blocks merken fr den Fall, da sein Folgeblock als nchstes hier bergeben wird
	getLastNameOfDirBlock (&loc.lastNameInPrevBlk, dirSector);
	loc.prevBlk = loc.currBlkNum;

	long totalEntries = loc.currStartIdx + entriesInBlk;
	if (blkDiff < (loc.dirTotalBlks-1)) {
		// Index fr Folgeblock eintragen
		dirLoc->dirBlk[blkDiff].firstIndex = totalEntries;
	} else {
		// Gesamteintrge eintragen, wenn es der letzte Dir-Block ist
		assert (dirLoc->dirRecCount == 0 || dirLoc->dirRecCount == totalEntries);
		dirLoc->dirRecCount = totalEntries;
	}
	
	// Apple-DirID ggf. nachtragen
	if (dirLoc->appleDirID == 0 && IsAppleDirID (dirID)) {
		dirLoc->appleDirID = dirID;
	}
}

void ClearLocBuffer (short volID)
// if volID==0, then reset all, otherwise clear only that volume's records
{
	if (LocBufSize > 0) {
		if (volID == 0) {
			LocBufUsed = 0;
		} else {
			// remove all records of this volume
			// Optimiertes Verfahren: Bentigt nur einen Durchgang, bewegt dabei
			// immer nur die zusammenhngenden, nicht zu lschenden Teilbereiche.
			Ptr dest, src, curr, end;
			curr = src = dest = LocBufStart;
			end = LocBufStart + LocBufUsed;
			long totalDiff = 0;
			long bytesToKeep = 0;
			long recLen;
			while (true) {
				LocArray *dirLoc = (LocArray*)curr;
				if (curr != end) recLen = LALEN (curr);
				if (curr == end || dirLoc->volID == volID) {
					// delete this record
					if (bytesToKeep > 0) {
						// delete all between dest and src
						if (src != dest) {
							BlockMoveData (src, dest, bytesToKeep);
						}
						dest += bytesToKeep;
						bytesToKeep = 0;
					}
					if (curr == end) break;	// exit the loop
					if ((Ptr)LocArrayLast == curr) LocArrayLast = nil;
					totalDiff += recLen;
				} else {
					// keep this record
					if (bytesToKeep == 0) {
						src = curr;	// here's the end of the records to be deleted
					}
					bytesToKeep += recLen;
				}
				curr += recLen;
			}
			LocBufUsed -= totalDiff;
			assert (LocBufUsed >= 0);
		}
		((LocArray*)(LocBufStart+LocBufUsed))->next = nil;	// Ende-Markierung
	}
}

void InitLocBuffer (Ptr start, long size)
{
	assert ((sizeof (LocArray) & 1) == 0);	// must have even size
	assert ((sizeof (LocEntry) & 1) == 0);	// must have even size
	
	LocBufStart = start;
	LocBufSize = size - sizeof (LocArray);
	LocBufEnd = start + LocBufSize;
	if (LocBufSize > 0) {
		((LocArray*)LocBufStart)->next = nil;	// Ende-Markierung
	}
	LocBufUsed = 0;
	LocArrayLast = nil;
}

// EOF
