#include <cctype>
#include <string.h>
#include "dcon.h"
#include "InternetConfig.h"
#include "PascalLib.h"
#include "FSM Utils.h"
#include "HFS CI Procs.h"
#include "FS API.h"
#include "FS Support.h"


long RoundDown (long value, long bound)
{
	long part = value % bound;
	return value - part;
}

long RoundUp (long value, long bound)
{
	long part = value % bound;
	if (part == 0) {
		return value;
	} else {
		return value - part + bound;
	}
}


Boolean HFSNamesMatch (const StringPtr name1, const StringPtr name2)
{
	return EqualString (name1, name2, false, true);	// HFS mode: caseSense=false,diacSense=true
}


/*
 * Support functions for file name parsing in "FSM Utils.cp"
 */

OSErr FFSGetParent (FSItemRef dirRef, ItemID *parID)
{
	FSItemRef parRef;
	OSErr err = FSItemGetParent (dirRef, &parRef);
	if (!err) {
		FSItemGetItemID (parRef, parID);
	} else if (err == fidNotFound) {
		ItemID id;
		FSItemGetItemID (dirRef, &id);
		if (id == fsRtDirID) {
			*parID = fsRtParID;
			err = noErr;
		}
	}
	return err;
}

OSErr FFSGetParent (ItemID dirID, ItemID &parID)
{
	if (dirID == fsRtDirID) {
		parID = fsRtParID;
		return noErr;
	}
	FSItemRef thisRef;
	OSErr err = FSItemLookup (dirID, &thisRef);
	if (err) {
		ErrorNum ("FSItemLookup in FFSGetParent", dirID);
	} else {
		err = FFSGetParent (thisRef, &parID);
		if (err) {
			ErrorNum ("FFSGetParent in FFSGetParent", dirID);
		}
		if (err == fidNotFound) err = dirNFErr;
	}
	return err;
}

OSErr FFSGetDir (Str31 name, ItemID parID, ItemID &destDirID)
{
	FSItemRef ref;
	OSErr err = FSItemLookup (parID, name, &ref);
	if (err) {
		ErrorNum ("FSItemLookup in FFSGetDir", parID);
	} else {
		if (FSItemIsDir (ref)) {
			FSItemGetItemID (ref, &destDirID);
		} else {
			ErrorNum ("FSItemIsDir in FFSGetDir", parID);
			err = dirNFErr;
		}
	}
	return err;
}


OSErr GetDirRef (Boolean highLevel, void *paramBlock, Boolean useIndex, DirEntryInfo &info)
// useIndex==2 means to count only files (excluding subdirs)
{
	OSErr				err;
	HFileInfo			*f = (HFileInfo*)paramBlock;
	ItemID				dirID;
	short				status;
	ObjectInfoRec		objectInfo;
	StringPtr			origNamePtr;

	/*
	 * Resolve the vol specification (in case it is a WD ref or the default dir)
	 */
	if (useIndex && f->ioFDirIndex) { origNamePtr = f->ioNamePtr; f->ioNamePtr = NULL; }
	err = GetCurrentDir ((ParmBlkPtr)paramBlock, (highLevel)?f->ioDirID:0, &dirID, &status);
	if (useIndex && f->ioFDirIndex) { f->ioNamePtr = origNamePtr; }
	if (err) return err;

	/*
	 * resolve the dir path if ioNamePtr is used
	 */
	if (!useIndex || f->ioFDirIndex == 0) {
		err = ParsePathName (dirID, f->ioNamePtr, &objectInfo);
		if (err) return err;
		dirID = objectInfo.dirID;
	}
	
	/*
	 * Find the ISO dir entry
	 */
	if (useIndex && f->ioFDirIndex < 0) {
		// return info about the dir 'd->ioDrDirID' (ioNamePtr will be ignored)
		if (dirID == fsRtParID) return fnfErr;
		err = FSItemLookup (dirID, &info.objRef);
		if (err) {
			ErrorNum ("FSItemLookup 1 in GetDirRef", dirID);
		}
		FSItemGetItemID (info.objRef, &info.objID);
		AssertNum (!err || err == fnfErr, err);
		info.objIdx = 0;
	} else if (useIndex && f->ioFDirIndex > 0) {
		// query the dir
		if (dirID == fsRtParID) return fnfErr;
		err = FSItemLookup (dirID, f->ioFDirIndex, useIndex==2, &info.objRef);
		if (err) {
			ErrorNum ("FSItemLookup 2 in GetDirRef", dirID);
		}
		FSItemGetItemID (info.objRef, &info.objID);
		AssertNum (!err || err == fnfErr, err);
		info.objIdx = f->ioFDirIndex;
	} else {
		// return the specified file/dir
		err = GetDirRecRefByName (dirID, objectInfo.fileName, info);
		AssertNum (!err || err == fnfErr, err);
	}
	info.parID = dirID;
	AssertNum (!err || err == fnfErr, err);
	return err;
}

OSErr GetDirRecRefByName (ItemID dirID, Str31 name, DirEntryInfo &info)
{
	OSErr err;
	if (dirID == fsRtParID) {
		// special case: volume name is referenced like a directory
		if (HFSNamesMatch (name, CurrVCB->vcbVN)) {
			err = FSItemLookup (fsRtDirID, &info.objRef);
			FSItemGetItemID (info.objRef, &info.objID);
			Assert (!err && info.objID == fsRtDirID);
		} else {
			err = fnfErr;
		}
	} else if (!name || !name[0]) {
		// empty name
		err = FSItemLookup (dirID, &info.objRef);
		if (err) {
			ErrorNum ("FSItemLookup 1 in GetDirRecRefByName", dirID);
		}
		FSItemGetItemID (info.objRef, &info.objID);
		AssertNum (!err || err == fnfErr, err);
	} else {
		err = FSItemLookup (dirID, name, &info.objRef);
		if (err) {
			ErrorNum ("FSItemLookup 2 in GetDirRecRefByName", err);
		}
		FSItemGetItemID (info.objRef, &info.objID);
		AssertNum (!err || err == fnfErr, err);
	}
	info.objIdx = 0;
	info.parID = dirID;
	return err;
}

OSErr GetNameOfDir (ItemID dirID, Str31 name)
{
	FSItemRef ref;
	OSErr err = FSItemLookup (dirID, &ref);
	if (err) {
		ErrorNum ("FSItemLookup in GetNameOfDir", dirID);
	}
	FSItemGetItemName31 (ref, name);
	assert (FSItemIsDir (ref));
	return err;
}

static	OSType	defaultType, defaultCreator;
static	ICInstance	icInst = nil;
static	Boolean	canUseIC = false;
static	Boolean	initedIC = false;

static Boolean isExtensionVar (StringPtr name, StringPtr extension)
// Returns true if name has a file extension that matches extension.
{
	short nameIndex, extensionIndex;
	short extLen = extension[0], nameLen = name[0];
	if (nameLen >= extLen) {
		nameIndex = (short)(nameLen - extLen + 1);
		extensionIndex = 1;
		while (extensionIndex <= extLen) {
			if (name[nameIndex] != toupper(extension[extensionIndex])) {
				return false;
			}
			++nameIndex;
			++extensionIndex;
		}
		return true;
	}
	return false;
}

static ICMapEntryPtr findMapping (StringPtr upName, const Ptr mapStart, const long mapSize)
// Loop through each of the entries in the handle looking for the longest match.
{
	short longestLenSoFar = 0, posOfLongestLenSoFar;
	long posIndex = 0;
	while ((posIndex+6) < mapSize) {
		ICMapEntryPtr me = (ICMapEntryPtr) (mapStart + posIndex);
		if (me->fixed_length != ICmap_fixed_length || me->total_length < ICmap_fixed_length) {
			// there's an error in the mapping table -> abort search
			return nil;
		}
		/* We've found a new longest match if:
			o not_incoming flag bit is clear
			o it's longer than the previous max
			o it's longer than the file name
			o it matches the last N chars of the filename
		*/
		if ((me->extension[0] > longestLenSoFar)
		&&  ((me->flags & ICmap_not_incoming_mask) == 0)
		&&  isExtensionVar (upName, me->extension))
		{
			// Record the new longest entry.
			longestLenSoFar = me->extension[0];
			posOfLongestLenSoFar = (short)posIndex;
		}
		posIndex += me->total_length;
	}
	if (longestLenSoFar > 0) {
		return (ICMapEntryPtr) (mapStart + posOfLongestLenSoFar);
	}
	return nil;
}

void GetDefaultTypeAndCreator (StringPtr fileName, OSType *type, OSType *creator)
{
	#if !GENERATINGPOWERPC	// +++ doesn't work with PPC yet
		if (canUseIC && fileName) {
			Ptr mappingPtr;
			long len;
			ICError err = ICGetMappingInterruptSafe (icInst, &mappingPtr, &len);
			if (!err) {
				Str31 upName;
				StringPtr s = fileName, d = upName;
				short len2 = *s++;
				if (len2 > 31) {
					#if TRACING
						DebugStr("\poops - len2 > 31!");
					#endif
					len2 = 31;
				}
				*d++ = len2;
				while (len2--) { *d++ = (Byte)toupper(*s++); }
				ICMapEntryPtr me = findMapping (upName, mappingPtr, len);
				if (me) {
					if (type)    *type    = me->file_type;
					if (creator) *creator = me->file_creator;
					return;
				}
			}
		}
	#endif
	if (type)    *type    = defaultType;
	if (creator) *creator = defaultCreator;
}

static long icGetActualAPIVersion ()
{
	ICError err;
	ComponentInstance componentInst;
	err = ICGetComponentInstance (icInst, &componentInst);
	if (!err) {
		return GetComponentVersion (componentInst);
	} else {
		return 0;
	}
}

void InitTypeAndCreatorTranslation ()
// Prepare Internet Config use for extension mapping
// must be called with A5 setup for a valid grafport!
{
	defaultType    = 'TEXT';
	defaultCreator = 'ttxt';
	
	#if !GENERATINGPOWERPC
		if (!initedIC) {
			ICError err = ICStart (&icInst, kCreator);
			if (icInst != NULL) {
				if (icGetActualAPIVersion () >= kICComponentInterfaceVersion3) {	// IC Version must be 2.0 or higher
					err = ICFindConfigFile (icInst, nil, 0);
					if (!err) {
						initedIC = true;
					}
				}
			}
		}
		if (initedIC && !canUseIC) {
			canUseIC = ICRequiresInterruptSafe (icInst) == noErr;
		}
	#endif
}

void ExitTypeAndCreatorTranslation ()
{
	canUseIC = false;
	initedIC = false;
	#if !GENERATINGPOWERPC
		if (icInst != NULL) {
			ICStop (icInst);
			icInst = NULL;
		}
	#endif
}


// EOF
