/*
 * main of FSM Client
 */

#include <string.h>
#include <stddef.h>

#include "PascalLib.h"
#include "Logging.h"

#include "FSM.h"
#include "FSM Main.h"
#include "FSM Utils.h"
#include "HFS CompInt.h"


const	long	kFSStackSize = 1024 * 24L;


typedef struct {	/* 'vers' resource definition, see Inside Mac docs */
	Byte	vm;		/* first part of version number in BCD */
	Byte	vn;		/* second and third part of version number in BCD */
	Byte	vt;		/* development: 0x20, alpha: 0x40, beta: 0x60, release: 0x80 */
	Byte	vd;		/* stage of prerelease version in BCD */
	short	region;	/* region code */
	char	str[];	/* version strings */
} VersionRecord;


static FSDCommUPP fsCommProc = nil;
static HFSCIUPP compInterfProc = nil;
static Lg2PhysUPP log2PhyProc = nil;


static Boolean UnmountFSVols (short fsID)
// returns false when it's not safe to de-install the File System because there are still volumes mounted by it.
{
	HVolumeParam pb;
	short err, idx;
	Str255 name;
	Boolean ok = true;

	idx = 1;
	do {
		pb.ioVolIndex = idx;
		pb.ioNamePtr = name;
		err = PBHGetVInfoSync ((HParamBlockRec*)&pb);
		if (err) break;
		if (pb.ioVFSID == fsID) {
			err = UnmountVol (nil, pb.ioVRefNum);
			#if !GENERATINGCFM && STANDALONE
				if (err) {
					pb.ioNamePtr = nil;
					err = PBHUnmountVol ((ParmBlkPtr)&pb);
					pb.ioNamePtr = name;
				}
			#endif
			if (!err) {
				--idx;
				Eject (nil, pb.ioVDrvInfo);
			} else {
				ok = false;
			}
		}
		++idx;
	} while (true);
	return ok;
}

static UniversalProcPtr gOldHandler = nil;
static	Ptr		gStackLo = nil;
static	Boolean	gFSInstalledByMe = false;


static void uninstallFS ()
{
	if (gOldHandler) {
		SetToolTrapAddress ((UniversalProcPtr)gOldHandler, _ExitToShell);
		gOldHandler = nil;
	}
	if (gFSInstalledByMe) {
		UnmountFSVols (kFSID);
	}
	if (fsCommProc) { DisposeRoutineDescriptor (fsCommProc); fsCommProc = nil; }
	if (compInterfProc) { DisposeRoutineDescriptor (compInterfProc); compInterfProc = nil; }
	if (log2PhyProc) { DisposeRoutineDescriptor (log2PhyProc); log2PhyProc = nil; }
	if (gFSInstalledByMe) {
		gFSInstalledByMe = false;
		RemoveFS (kFSID);
	}
	UnloadIcons ();
	// ExitTypeAndCreatorTranslation ();
	if (gStackLo) { DisposePtr (gStackLo); gStackLo = nil; }
	LogExit ();
}

static pascal void CleanupAndExit (void)
{
	uninstallFS ();
	ExitToShell ();
}

static void InitQuit ()
{
	if (gOldHandler == nil) {
		gOldHandler = NGetTrapAddress (_ExitToShell, 1);
		SetToolTrapAddress (NewRoutineDescriptor ((ProcPtr)CleanupAndExit, kPascalStackBased, GetCurrentISA()), _ExitToShell);
	}
}

enum
{
	/* set if FFS' log2PhyProc maps to physical disk blocks as documented */
	hfsCIL2PCanMapFileBlocksBit = 15,
	hfsCIL2PCanMapFileBlocksMask= 0x00008000
};


static OSErr InstallMyFS (Ptr stackLo, long stackSize, StringPtr fsName, short fsID)
{
	FSDRec	fsdRec;
	OSErr	err;
	short	recSize = sizeof (FSDRec);
	
	memset (&fsdRec, 0, recSize);
	
	fsdRec.fsdLength = recSize;
	fsdRec.fsdVersion = fsdVersion1;
	fsdRec.fileSystemFSID = fsID;
	pstrcpy (fsdRec.fileSystemName, fsName);
	fsdRec.fileSystemGlobalsPtr = (Ptr)(long)0xB5559999;	// some bus error causing address
	fsdRec.fileSystemCommProc = fsCommProc;

	err = InstallFS (&fsdRec);
	if (err) return err;
	
	gMountEnableFlags = 1;
	gFSInstalledByMe = true;
	
	err = GetFSInfo (fsmGetFSInfoByFSID, fsID, &recSize, &fsdRec);

	if (!err) {	
		fsdRec.fsdHFSCI.compInterfMask = hfsCIHasHLL2PProcMask | hfsCIDoesDynamicLoadMask | hfsCIDoesHFSMask | hfsCIL2PCanMapFileBlocksMask;
		fsdRec.fsdHFSCI.compInterfProc = compInterfProc;
		fsdRec.fsdHFSCI.log2PhyProc = log2PhyProc;
		fsdRec.fsdHFSCI.stackTop = stackLo+stackSize;
		fsdRec.fsdHFSCI.stackSize = stackSize;
		fsdRec.fsdHFSCI.stackPtr = NULL;
		fsdRec.fsdHFSCI.idSector = -1;	// 512Byte-Sector number or -1 for none
	
		err = SetFSInfo (fsID, sizeof (FSDRec), &fsdRec);
	}
	
	if (err) {
		uninstallFS ();
	}
	
	return err;
}

static void error (short errCode, short *retCode, StringPtr msg)
{
	*retCode = errCode;
	LogFStr (LKFatal, HERE, "error (%d)", errCode);
	GetIndString (msg, 128, errCode);
	uninstallFS ();
}

Boolean FSMClientInstall (short *errCode, Str255 errMsg, Boolean makeResident)
{
	OSErr	err;
	Handle	h;

	errMsg[0] = 0;
	
	LogInit ("FS");
	Note ("Installing FSM Client");

	if (HasRequiredFSM (&fsmVersion) != 0) {
		error (2, errCode, errMsg);
		return false;
	}

	if (!makeResident) {
		InitQuit ();
	}
	
	{
		FSDRec	fsdRec;
		short	recSize = sizeof (FSDRec);
		err = GetFSInfo (fsmGetFSInfoByFSID, kFSID, &recSize, &fsdRec);
		if (err == noErr) {	
			error (1, errCode, errMsg);
			return false;
		}
	}

	// version aus vers-Rsrc holen
	h = Get1Resource ('vers', 1);
	if (h) {
		gFFSVersion = *(long*)*h;
		ReleaseResource (h);
	}
	
	gStackLo = NewPtrSys (kFSStackSize);
	if (gStackLo == NULL) {
		Error ("NewPtrSys(kFSStackSize) failed.");
		error (3, errCode, errMsg);
		return false;
	}

	THz origZone = GetZone();
	SetZone (SystemZone ());
	{
		fsCommProc = NewFSDCommProc (FSCommProc);
		compInterfProc = NewHFSCIProc (FSHFSCIProc);
		log2PhyProc = NewLg2PhysProc (FSLg2PhysProc);
		if (!fsCommProc || !compInterfProc || !log2PhyProc) {
			Error ("New...Proc() failed.");
			error (3, errCode, errMsg);
			return false;
		}

		err = InstallMyFS (gStackLo, GetPtrSize (gStackLo), kFSName, kFSID);
		if (err) {
			LogFStr (LKFatal, NULL, "Installation failed: %d", (int) err);
			error (4, errCode, errMsg);
			return false;
		}

		LoadIcons ();
	}
	SetZone (origZone);
	
	InformFSM (fsmDrvQElChangedMessage, NULL);

	*errCode = 0;
	return true;
}

UInt16 FSMClientGetFSID ()
{
	return kFSID;
}

Boolean FSMClientUninstall ()
{
	if (!UnmountFSVols (kFSID)) {
		return false;
	}
	uninstallFS ();
	return true;
}

// EOF
