/*
 * HFS CompInt.c
 *
 * HFS Component Interface
 */

#include "FSM.h"
#include "FSM Client Msgs.h"
#include "HFS CompInt.h"
#include "HFS CI Procs.h"
#include "FSM Utils.h"
#include "FSM Main.h"
#include "FS Support.h"

// prototypes for Dispatch Table
static OSErr DoParamErr (Boolean highLevel, void *paramBlock);
static OSErr DoWPrErr (Boolean highLevel, void *paramBlock);
static OSErr DoNoErr (Boolean highLevel, void *paramBlock);

typedef OSErr (*HFSProcPtr) (Boolean highLevel, void *paramBlock);

UInt32	fsmVersion, gFFSVersion;
UInt16	gMountEnableFlags;
SInt16	gDriveToMount;
Ptr		gFSPrivateDataPtr = NULL;


const short	kNumFileManagerProcs = 70; // FlushFile is 0xA045

const HFSProcPtr	kFileManagerDispatchTable[kNumFileManagerProcs] = {
				DoOpen,			
				DoClose,
				DoRead,
				DoWPrErr,			// Write
				NULL,
				NULL,
				NULL,
				DoGetVolInfo,
				DoWPrErr,			// Create
				DoWPrErr,			// Delete
// 10				
				DoOpenRF,
				DoWPrErr,			// Rename
				DoGetFInfo,
				DoWPrErr,			// SetFileInfo
				DoUnmountVol,
				DoMountVol,
				DoWPrErr,			// Allocate
				DoGetEOF,
				DoWPrErr,			// SetEOF
				DoNoErr,			// FlushVol
// 20				
				DoParamErr,			// GetVol
				DoSetVol,
				NULL,
				DoEject,
				DoGetFPos,			// 0xA018 (24)
				NULL, NULL, NULL, NULL, NULL,	// 25 - 29
// 30
				NULL, NULL, NULL, NULL, NULL,	// 30 - 34
				NULL, NULL, NULL, NULL, NULL,	// 35 - 39
// 40
				NULL, NULL, NULL, NULL, NULL,	// 40 - 44
				NULL, NULL, NULL, NULL, NULL,	// 45 - 49
// 50
				NULL, NULL, NULL, 	 			// 50 - 52
				DoOffline,
				NULL, NULL, NULL, NULL, NULL,	// 54 - 58
				NULL,
// 60
				NULL, NULL, NULL, NULL, NULL,	// 60 - 64
				DoNoErr,			// SetFilLock
				DoNoErr,			// RstFilLock
				DoParamErr,			// SetFilType
				DoSetFPos,
				DoNoErr				// FlushFile
// 70
};				
				

const short	kNumHFSProcs = 49;

const HFSProcPtr	kHFSDispatchTable[kNumHFSProcs] = {
				NULL,
				DoOpenWD,
				DoCloseWD,
				NULL,
				NULL,
				DoWPrErr,			// CatMove
				DoWPrErr,			// DirCreate
				DoGetWDInfo,
				DoGetFCBInfo,
				DoGetCatInfo,
// 10
				NULL, NULL, NULL, NULL, NULL,
				NULL,
				DoNoErr,			// LockRange
				DoNoErr,			// UnlockRange
				NULL, NULL,
// 20
				DoParamErr,			// DoCreateFileIDRef,
				DoParamErr,			// DoDeleteFileIDRef,
				DoResolveFileIDRef,
				DoParamErr,			// ExchangeFiles
				DoCatSearch,
				NULL,
				DoParamErr,			// OpenDF (never passed)
				DoMakeFSSpec,
				NULL,
				NULL,
// 30	
				NULL,
				NULL,
				DoDTGetPath,		// DTGetPath
				DoDTCloseDown,		// DTCloseDown
				DoParamErr,			// DTAddIcon
				DoDTGetIcon,
				DoDTGetIconInfo,
				DoParamErr,			// DTAddAPPL
				DoParamErr,			// DTRemoveAPPL
				DoDTGetAPPL,
// 40
				DoParamErr,			// DTSetComment
				DoParamErr,			// DTRemoveComment
				DoDTGetComment,
				DoNoErr,			// DTFlush
				DoNoErr,			// DTReset
				DoDTGetInfo,		// DTGetInfo
				DoParamErr,			// DTOpenInform
				DoParamErr,			// DTDelete
// 48
				DoGetVolParms,
};			



static OSErr DoParamErr (Boolean highLevel, void *paramBlock)
{
#pragma unused (highLevel, paramBlock)
	return paramErr;
}							

static OSErr DoWPrErr (Boolean highLevel, void *paramBlock)
{
#pragma unused (highLevel, paramBlock)
	return wPrErr;
}							

static OSErr DoNoErr (Boolean highLevel, void *paramBlock)
{
#pragma unused (highLevel, paramBlock)
	return noErr;
}							


/*
 * Main entry to all toolbox & system calls
 */
pascal OSErr FSHFSCIProc
	(VCBPtr theVCB, short selectCode, void *paramBlock, void *fsdGlobalPtr, short fsid)
{
#pragma unused (fsid)

	OSErr			result = paramErr;
	unsigned short	procIndex, tableSize;
	const HFSProcPtr		*dispatchTable;
	HFSProcPtr		hfsProc;
	long			oldA5;

	oldA5 = SetA5 ((long) fsdGlobalPtr);
	Entering (theVCB);

	LogFStr (LKUser10, HERE, "HFSCIProc, calling selectCode: %04X", selectCode);

	if ((selectCode & 0xF000) == 0xA000) {
		dispatchTable = kFileManagerDispatchTable;
		tableSize = kNumFileManagerProcs;
	} else {
		dispatchTable = kHFSDispatchTable;
		tableSize = kNumHFSProcs;
	}

	procIndex = selectCode & 0x00FF;	// clear the trap attribute bits (async etc.)
	if (procIndex < tableSize) {
		hfsProc = dispatchTable[procIndex];
	} else {
		hfsProc = NULL;
	}

	#if TRACING
		short freebufs = FreeSectorBuffers ();
	#endif
	
	if (hfsProc != NULL) {
		Boolean isHighLevel = !((selectCode & 0xF200) == 0xA000);
		result = (*hfsProc) (isHighLevel, paramBlock);
	} else {
		result = paramErr;
	}

	#if TRACING
		short freebufs2 = FreeSectorBuffers ();
		if (freebufs2 != freebufs) {
			LogFStr (LKError, HERE, "HFSCIProc %04X: free was %d, now %d", selectCode, freebufs, freebufs2);
		}
	#endif

	LogFStr (LKUser10, HERE, "HFSCIProc, returned with err = %d", result);

	Exiting ();
	SetA5 (oldA5);
	return result;
}

pascal OSErr FSExtendFileProc (VCBPtr theVCB, FCBRecPtr theFCB, short options, long *reqCount)
{
#pragma unused (theVCB, theFCB, options, reqCount)
	// disk is read only
	return dskFulErr;
}


#if 1
// High-Level version
pascal OSErr FSLg2PhysProc (void *fsdGlobalPtr, VCBPtr theVCB, FCBRecPtr theFCB, 
							short fileRefNum, unsigned long filePosition, unsigned long reqCount,
							unsigned long *volOffset, long *contiguousBytes)
{
	OSErr	err;
	long	oldA5;
	oldA5 = SetA5 ((long) fsdGlobalPtr);
	Entering (theVCB);
	err = DoLg2Phys (theVCB, theFCB, fileRefNum, filePosition, reqCount, volOffset, contiguousBytes);
	Exiting ();
	SetA5 (oldA5);
	return err;
}
#else
/*
// Low-Level version
pascal asm OSErr FSLg2PhysProc (void *fsdGlobalPtr, VCBPtr theVCB, FCBRecPtr theFCB, 
							short fileRefNum, unsigned long filePosition, unsigned long reqCount,
							unsigned long *volOffset, long *contiguousBytes)
{
	//  A2: Ptr auf VCB-Entry des Volumes, auf dem das File liegt.
	//  A1: Ptr auf Start des FCB-Puffers
	//  D1.L: File-Handle als Index in FCB-Puffer
	//  D5.L: Offset
	//  D4.L: Lnge (sinnvollerweise immer Dateilnge minus Offset)
	// returns:
	//  D6.L: Lnge
	//  D3.L: Block-Nr (Drive-Offset / 512)
	movem.l	d1/d2/a0/a1/a5,-(a7)
	lea		0xB5555555,a5	// signals use of wrong a5
	move.l	a2,-(a7)
	pea		0(a1,d1.w)
	move.w	d1,-(a7)
	...
	movem.l	(a7)+,d1/d2/a0/a1/a5
	tst.w	d0
	rts
}
*/
#endif


static OSErr DoMessage (tSampleFSMessage *pb)
{
	if (Assert (pb->version == 1 && pb->msgSize >= 16)) {
		switch (pb->message) {

			case kSampleFSMsgGetVers:
				if (pb->msgSize >= 20) {
					pb->ffsVersion = gFFSVersion;
					pb->msgSize = 20;
					return noErr;
				}
				break;

			case kSampleFSMsgEnquireStatus:
				if (pb->msgSize >= 18) {
					pb->enable = gMountEnableFlags;
					pb->msgSize = 18;
					return noErr;
				}
				break;

			case kSampleFSMsgEnable:
				if (pb->msgSize >= 18) {
					gMountEnableFlags = pb->enable;
					return noErr;
				}
				break;

			case kSampleFSMsgSetDrive:
				if (pb->msgSize >= 18) {
					gDriveToMount = pb->driveNum;
					LogPStr (LKUser11, HERE, "\pmounting announced for drive ^0", gDriveToMount);
					return noErr;
				}
				break;

			case kSampleFSMsgUninstall:
				if (pb->msgSize >= 16) {
					Boolean ok = FSMClientUninstall ();
					return ok ? noErr : -1;
				}
				break;

			case kSampleFSMsgPassFSDataPtr:
				if (pb->msgSize >= 20) {
					gFSPrivateDataPtr = pb->ptr;
					return noErr;
				}
				break;
		}
	}
	return paramErr;
}


static OSErr enableHFSCI ()
{
	FSDRec	fsdRec;
	OSErr	err;
	short	recSize = sizeof (FSDRec);
	err = GetFSInfo (fsmGetFSInfoByFSID, kFSID, &recSize, &fsdRec);
	if (err == noErr) {
		fsdRec.fsdHFSCI.compInterfMask |= hfsCIResourceLoadedMask | fsmComponentEnableMask;
		err = SetFSInfo (kFSID, sizeof (FSDRec), &fsdRec);
	}
	return err;
}

static OSErr disableHFSCI ()
{
	FSDRec	fsdRec;
	OSErr	err;
	short	recSize = sizeof (FSDRec);
	err = GetFSInfo (fsmGetFSInfoByFSID, kFSID, &recSize, &fsdRec);
	if (err == noErr) {
		fsdRec.fsdHFSCI.compInterfMask &= ~ (hfsCIResourceLoadedMask | fsmComponentEnableMask);
		err = SetFSInfo (kFSID, sizeof (FSDRec), &fsdRec);
	}
	return err;
}

pascal OSErr FSCommProc (short message, void *paramBlock, void *fsdGlobalPtr)
{
	long	oldA5;
	OSErr	err = paramErr;

	oldA5 = SetA5 ((long) fsdGlobalPtr);

	Mark ("FSCommProc enter");

	switch (message) {
		case ffsNopMessage:				/* nop, should always return noErr */
			err = noErr;
			break;

		case ffsGetIconMessage:			/* return disk icon and mask */
			err = DoGetIcon ((FSMGetIconRec *) paramBlock);
			break;

		case ffsIDDiskMessage:			/* identify the about-to-be-mounted volume */
			Note ("IDDisk");
			err = DoIDDisk ((Ptr) paramBlock);
			break;

		case ffsLoadMessage:			/* load in the FFS */
			{
				THz origZone = GetZone();
				SetZone (SystemZone ());
				long savedA5 = SetCurrentA5();
				InitTypeAndCreatorTranslation ();
				SetA5 (savedA5);
				SetZone (origZone);
			}
			enableHFSCI ();
			err = noErr;
			break;

		case ffsUnloadMessage:			/* unload the FFS */
			ExitTypeAndCreatorTranslation ();
			disableHFSCI ();
			err = noErr;
			break;

		case ffsIDVolMountMessage:		/* identify a VolMountInfo record */
			Note ("IDVolMount");
			err = DoIDVolMount ((VolumeMountInfoHeader *) paramBlock);
			break;

		case ffsInformMessage:			/* FFS defined message */
			err = DoMessage ((tSampleFSMessage*) paramBlock);
			break;

		case ffsGetIconInfoMessage:		// not used due to DTS
			assert (false);
			err = paramErr;
			break;

		default:
			err = paramErr;
	}

	Mark ("FSCommProc exit");
	
	SetA5 (oldA5);
	return err;
}

// EOF
