/*
	
	FSAPI.cp
	
*/
#pragma once
#include "FS API.h"
#include "ResForkInfo.h"
#include "PascalLib.h"
#include <Notification.h>
#include <DriverServices.h>

static tResForkInfoPtr sRFIRootInfo = NULL;

void FSAtomicStart (VCBPtr currVCB, Ptr privateDataPtr)
{
#pragma unused(currVCB)
	// attn: currVCB may be NULL!

	sRFIRootInfo = tResForkInfoPtr(privateDataPtr);
}

void FSAtomicStop ()
{
	// NOP, as our data is always locked
}

/*
 * Item (file/dir) lookup
 */


static bool ItemIsValid(tRFINodeInfoPtr node)
{
	if (node == NULL)
		return false;

	switch (node->nodeType)
	{
		case kRFIDirectoryType:
		case kRFIResourceType:
			return true;
			
		default:
			return false;
	}
}

static tRFINodeInfoPtr RefToNode(FSItemRef ref)
{
	return ItemIsValid(tRFINodeInfoPtr(ref)) ? tRFINodeInfoPtr(ref) : NULL;
}

OSErr FSItemLookup (ItemID itemID, FSItemRef *hdl)
{
	if (itemID == fsRtDirID)
		itemID = ItemID(sRFIRootInfo);

	*hdl = (FSItemRef)(itemID);
	return ItemIsValid(tRFINodeInfoPtr(itemID)) ? noErr : fidNotFound;
}

OSErr FSItemLookup (ItemID parID, UInt32 index, Boolean filesOnly, FSItemRef *hdl)
{
	if (parID == fsRtDirID)
		parID = (ItemID)sRFIRootInfo;

	if (not ItemIsValid(tRFINodeInfoPtr(parID)))
		return dirNFErr;

	tRFIDirectoryPtr dir = tRFIDirectoryPtr(RefToNode(FSItemRef(parID)));
		
	dir++;
	
	int i = 0;
	int numItems = dir->numItems;
	while (i < index)
	{
		dir = tRFIDirectoryPtr(Ptr(dir) + dir->nodeInfo.nodeSize);
		numItems--;
			
		if (not filesOnly or (dir->nodeInfo.nodeType != kRFIDirectoryType))
		{
			i++;
		}
		
		if (numItems == 0)
		{
			return fnfErr;
		}
	}

	*hdl = (FSItemRef)dir;

	return noErr;
}

OSErr FSItemCountItems (FSItemRef hdl, Boolean filesOnly, UInt32 *items)
{
	tRFIDirectoryPtr dir = tRFIDirectoryPtr(RefToNode(hdl));
	
	if (dir == NULL)
		return dirNFErr;
	
	if (filesOnly)
	{
		*items = dir->numItems;
	}
	else
	{
		*items = 0;

		dir++;
		
		for (int i = 0; i < dir->numItems; i++)
		{
			if (dir->nodeInfo.nodeType != kRFIResourceType)
			{
				*items++;
			}

			dir = tRFIDirectoryPtr(Ptr(dir) + dir->nodeInfo.nodeSize);
		}
	}

	return noErr;
}

OSErr FSItemLookup (ItemID parID, StringPtr name, FSItemRef *hdl)
{
	if (not ItemIsValid(tRFINodeInfoPtr(parID)))
		return dirNFErr;

	tRFIDirectoryPtr dir = tRFIDirectoryPtr(parID);

	dir++;
	
	for (int i = 0; i < dir->numItems; i++)
	{
		if (!pstrcmp(name, dir->nodeInfo.nodeName))
		{
			*hdl = (FSItemRef)dir;
			return noErr;
		}

		dir = tRFIDirectoryPtr(Ptr(dir) + dir->nodeInfo.nodeSize);
	}
	
	return fnfErr;
}

OSErr FSItemGetParent (FSItemRef hdl, FSItemRef *parHdl)
{
	if (not ItemIsValid(tRFINodeInfoPtr(hdl)))
		return fnfErr;

	tRFINodeInfoPtr node = tRFINodeInfoPtr(hdl);
	
	if (node->parent != NULL)
	{
		*parHdl = FSItemRef(node->parent);
		return noErr;
	}
	else
	{
		return dirNFErr;
	}
}

/*
 * Item access
 */

Boolean FSItemIsDir (FSItemRef hdl)
{
	if (not ItemIsValid(tRFINodeInfoPtr(hdl)))
		return fnfErr;

	tRFINodeInfoPtr node = tRFINodeInfoPtr(hdl);
	
	return (node->nodeType == kRFIDirectoryType);
}

void FSItemGetItemID (FSItemRef hdl, ItemID /*out*/ *itemID)
{
	if (hdl == sRFIRootInfo)
		*itemID = fsRtDirID;
	else
		*itemID = ItemID(hdl);
}

Boolean FSItemGetItemName31 (FSItemRef hdl, Str31 /*out*/ itemName)
{
	// returns false if name is > 31 chars and if the FS has no good way to shorten it
	
	tRFINodeInfoPtr node = tRFINodeInfoPtr(hdl);
	
	if (*node->nodeName <= 31)
	{
		pstrcpy(itemName, node->nodeName);
		return true;
	}
	else
	{
		return false;
	}
}

void FSItemGetItemNameFull (FSItemRef hdl, Str255 /*out*/ itemName)
{
	tRFINodeInfoPtr node = tRFINodeInfoPtr(hdl);
	pstrcpy(itemName, node->nodeName);
}

void FSItemGetItemProperties (FSItemRef hdl, FSItemProperties /*out*/ *prop)
{
	tRFIResourcePtr node = tRFIResourcePtr(RefToNode(hdl));

	prop->isDir = FSItemIsDir(hdl);

	if (not prop->isDir)
	{
		prop->dataForkLogLen = node->resSize;
		prop->dataForkPhysLen = (node->resSize + 511) & ~512;
		prop->rsrcForkLogLen = prop->rsrcForkPhysLen = 0;
	}

	prop->crDate = sRFIRootInfo->crDate;
	prop->mdDate = sRFIRootInfo->mdDate;
	prop->isReadOnly = true;
	prop->isHidden = false;
	prop->reserved = 0;
}

OSErr FSItemFileOpen (FSItemRef hdl, Boolean rsrcFork, FSOpenFileRecord /*out*/ *ref)
{
	if (not ItemIsValid(tRFINodeInfoPtr(hdl)))
		return fnfErr;

	if (rsrcFork)
		return errFSForkNotFound;

	ref->x[0] = (long)hdl;

	return noErr;
}

OSErr FSItemFileRead (FSOpenFileRecPtr ref, UInt32 fileOfs, UInt32 numBytes, Ptr destAddr)
{
	//+++
	
}

void FSItemFileClose (FSOpenFileRecPtr /*ref*/)
{
	// NOP, we don't close anything
}

/*
 * Volume mounting
 */

UInt32 FSPrivateDataSize ()
{
	// must be a constant value!
	return sizeof(tRFINodeInfoPtr);
}

OSErr FSFinishMount (VCBPtr /*vcb*/)
{
	// NOP
	
	return noErr;
}

// TaskTime stuff:

typedef void (*TaskTimeProc) (UInt32 param);

struct TaskTimeInfo {
	NMRec			nmRec;
	TaskTimeProc	proc;
	UInt32			param;
};
static TaskTimeInfo	cTaskTimeInfo;

static 	pascal void 		NMResponse(NMRec *nmRec);
static 	RoutineDescriptor	cNMResponseRD = BUILD_ROUTINE_DESCRIPTOR(uppNMProcInfo, NMResponse);

// NMResponse is called by the Notification Manager at task time. All we have to do is
// remove the notification record, reset the busy flag, and call the user's proc.
// CAUTION: Doing it that way allows clients to post another request before the first one
// has completed.

static bool cIsBusy = false;

pascal void NMResponse(NMRec *nmRec)
{
	TaskTimeProc	proc = cTaskTimeInfo.proc;
	UInt32			param = cTaskTimeInfo.param;

	::NMRemove(nmRec);
	cIsBusy = false;

	(*proc) (param);
}


static OSStatus CallBackAtTaskTime(TaskTimeProc proc, UInt32 param)
{
	if (cIsBusy) // this might lock up, should use CompareAndSwap...
		return paramErr; 

	cIsBusy = true;
	
	::BlockZero(&cTaskTimeInfo, sizeof(TaskTimeInfo));

	cTaskTimeInfo.proc = proc;
	cTaskTimeInfo.param = param;
	
	cTaskTimeInfo.nmRec.qType = nmType;
	cTaskTimeInfo.nmRec.nmResp = &cNMResponseRD;

	OSStatus status = ::NMInstall(&cTaskTimeInfo.nmRec);

	if (status != noErr)
		cIsBusy = false;
		
	return status;
}

static void MyUnmountVolProc(UInt32 param)
{
	// This routine gets called back at task time by the notification manager.
	// We use it to close the file and to dispose of the ResForkInfo pointer.
	tResForkInfoPtr rfi = tResForkInfoPtr(param);
	
	::FSClose(rfi->rfRefNum);
	::DisposePtr(Ptr(rfi));
}

void FSUnmountVol ()
{
	CallBackAtTaskTime(MyUnmountVolProc, UInt32(sRFIRootInfo));
}

// EOF

