/* shutdown.c */

#include "AEADebugging.h"

// Standard C
#include <stdio.h>
#include <string.h>

// Mac Universal Headers
#include <Errors.h>
#include <Events.h>
#include <Processes.h>
#include <ShutDown.h>

// Apple Event Assistant
#include "AEAAppleEventSender.hh"
#include "AEA1AEIdleFunction.hh"

#define NAM

#ifdef NAM
extern "C" int shutdown_main(int argc, char *argv[]);
#else
#define shutdown_main main
#endif

static OSErr QuitProcesses(void);
static OSErr QuitPSN(const ProcessSerialNumber &inPSN);

int shutdown_main(int argc, char *argv[])
{
	if (argc < 2) {
		fprintf(stderr, "shutdown: %s\n", "missing option");
		return 1;
	}
	if (QuitProcesses() == noErr)
		if (strcmp(argv[1], "-h") == 0)
			ShutDwnPower();
		else if (strcmp(argv[1], "-r") == 0)
			ShutDwnStart();
		else
			DebugBeep();
	return -1;
}

OSErr QuitPSN(const ProcessSerialNumber &inPSN)
{
	AEAAppleEventSender quit(kCoreEventClass, kAEQuitApplication);
	quit.SetGizmoBuildString("");
	quit.SetAddress(inPSN);
	quit.Send();
	
	return noErr;
}

OSErr QuitProcesses(void)
{
	enum {kShutdownGracePeriod = 10 * 60};
	OSErr err;
	const ProcessSerialNumber psnNone = {0, kNoProcess};
	ProcessSerialNumber psn, psnMe, psnFinder, psnFSE;
	ProcessInfoRec procInfo;
	Boolean same;
	Boolean runningFinder = false, runningFSE = false, runningApps = false;
	Boolean waiting = false;
	enum {apps, fse, finder, done} quitting;
	
	err = GetCurrentProcess(&psnMe);
	if (err != noErr) return err;
	
	procInfo.processInfoLength = sizeof procInfo;
	procInfo.processName = NULL;
	procInfo.processAppSpec = NULL;
	
	quitting = apps;
	
	while (quitting != done) {
		EventRecord evt;
		long tcQuit, tcNow;
		
		psn = psnNone;
		runningFinder = runningFSE = runningApps = false;
		// loop over each process
		while (1) {
			//OSErr err;
			
			err = GetNextProcess(&psn);
			if (err == procNotFound) break; // Okay, that's everyone
			if (err != noErr) return err; // This shouldn't fail
			
			err = SameProcess(&psn, &psnMe, &same);
			if (err != noErr) return err; // This shouldn't fail
			if (same) continue; // Don't quit self yet
			
			err = GetProcessInformation(&psn, &procInfo);
			if (err != noErr) return err; // This shouldn't fail
			
			switch (procInfo.processSignature) {
				case 'MWDB':
					// don't try to quit the debugger (if we're running in it)
					break;
				case 'MACS':
					runningFinder = true;
					//psnFinder = psn;
					if (!waiting && quitting == finder) {
						err = QuitPSN(psn);
					}
					break;
				case 'hhgg':
					runningFSE = true;
					//psnFSE = psn;
					if (!waiting && quitting == fse) {
						err = QuitPSN(psn);
					}
					break;
				default:
					runningApps = true;
					char sig[5] = "----";
					*(OSType *)sig = procInfo.processSignature;
					fprintf(stderr, waiting ? "%s still running\n" : "quit %s\n", sig);
					fflush(stderr);
					if (!waiting) {
						err = QuitPSN(psn);
						(void)WaitNextEvent(nullEvent, &evt, 30, NULL);
					}
					break;
			}
		}
		tcNow = TickCount();
		if (!waiting)
			tcQuit = tcNow;
		if (quitting == finder)
			if (!runningFinder)
				quitting = done, waiting = false;
			else if (tcNow - tcQuit > kShutdownGracePeriod)
				quitting = done;
			else
				waiting = true;
		if (quitting == fse)
			if (!runningFSE)
				quitting = (runningFinder ? finder : done), waiting = false;
			else if (tcNow - tcQuit > kShutdownGracePeriod)
				quitting = done;
			else
				waiting = true;
		if (quitting == apps)
			if (!runningApps)
				quitting = (runningFSE ? fse : runningFinder ? finder : done), waiting = false;
			else if (tcNow - tcQuit > kShutdownGracePeriod)
				quitting = done;
			else
				waiting = true;
		FlushEvents(everyEvent, 0);
		
		if (waiting)
			(void)WaitNextEvent(nullEvent, &evt, kShutdownGracePeriod - (tcNow - tcQuit), NULL); // give apps a chance to quit
	}
	
	return (runningApps || runningFSE || runningFinder) ? userCanceledErr : noErr;
}
