// =================================================================================
//	FilterTable.h	 			1998 Sustainable Softworks. All rights reserved.
// =================================================================================
//	IP Filter Table
//	This module encapsulates the IP filter table
//	as a set of cover routines to hide the internal structure
//	and algorithm used to access the table.
//
//	Since this module executes in the STREAMs context, instance data is
//	allocated when the STREAM is opened rather than by instantiating
//	a C++ object.  For this reason, each of the module functions is
//	passed a pointer to its instance data.

#ifndef _H_FilterTable
#define _H_FilterTable
#pragma once

#include <types.h>
#include "MyTypes.h"

// flags for Filter entry
#define kFlagDirectionRead	0x01	// 0=write, 1=read
#define kFlagActionReject	0x02	// 0=pass,	1=reject
#define kFlagActionNoDial	0x04	// 0=nop,	1=no dial
#define kFlagAck			0x10
#define kFlagNoAck			0x20
#define kFlagAll			0x0133
#define kFlagWamnet			0x0100

// Values for protocol
#define kProtocolAny		0
#define kProtocolICMP		1
#define kProtocolTCP		6
#define kProtocolUDP		17
#define kProtocolGRE		47

// What to do with datagram
#define kFilterPass			0
#define kFilterReject		1
#define kFilterNoDial		2

// size of table: 100 entries from 1-100
#define kFilterTableDim		101

// The first entry of the Table (index 0) is reserved for bookkeeping 
struct FilterTableData {
	UInt16	lastEntry;		// index of last physical entry in table
	UInt16	lastUsed;		// index of last logical entry in table
	UInt16	firstActive;	// start of active list
	UInt16	firstDeleted;	// start of deleted list
	UInt32	ipAddress;		// dynamic IP address to use
};
typedef struct FilterTableData FilterTableData_t;

// Design Notes
//	For efficiency, the filter table is threaded by a series of links
//	using the entryID.  The entryID of each table entry points to
//	the next entry in the thread or zero for end of list.
//	Two threads are kept, one for active entries, and another for
//	deleted entries waiting to be re-used.
//
//	When a found entry is returned, the entryID parameter is the index
//	of that entry.  This index allows specific entries to be modified
//	without searching.



// ---------------------------------------------------------------------------
// forward function declarations [external]
// ---------------------------------------------------------------------------

void InitFilterTable(FilterEntry_t table[]);
// Initialize the filter table to be empty.

UInt8 FilterDatagram(FilterEntry_t table[], FilterEntry_t* filterData);
// Filter datagram based on table entries
//   Flags specifies whether it is a read or write side datagram
// Returns kFilterPass, kFilterReject, or kFilterNoDial.

UInt16 GetFilterEntry(FilterEntry_t table[]);
// Return index of unused entry or 0 if table is full

UInt16 FindFilterEntry(FilterEntry_t table[], FilterEntry_t* filterData);
// Return index of table entry with matching data:

Boolean WriteFilterEntry(FilterEntry_t table[], FilterEntry_t* filterData);
// To load a new entry, use entryID=0.  Returns false if table is full.

void DeleteFilterEntry(FilterEntry_t table[], UInt16 index, UInt16 prev);


#endif