/*
 * Copyright (c) 1998-2000 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 *
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */
/*
 * MPEnabler - An example of patching funtions in the Mac OS X Kernel
 *
 * Yes, I know that I said you don't patch functions in kernel,
 * but you should remember what your mother used to say,
 *
 *   "Do and I say, not as I do."
 * 
 * Josh de Cesare, MacHack 2000
 */

#include <mach/boolean.h>
#include <mach/policy.h>
#include <mach/processor_info.h>
#include <mach/vm_param.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/proc.h>
#include <sys/file.h>
#include <sys/vnode.h>
#include <sys/unistd.h>
#include <sys/buf.h>
#include <sys/ioctl.h>
#include <sys/tty.h>
#include <sys/disklabel.h>
#include <sys/vm.h>
#include <sys/sysctl.h>
#include <mach/mach_types.h>
#include <kern/task.h>
#include <vm/vm_kern.h>



#include "PatchIt.h"

kern_return_t MPEnablerStart(struct kmod_info *ki, void *data);
kern_return_t MPEnablerStop(struct kmod_info *ki, void *data);

void SetICTC(long on, long cycles);


extern kern_return_t processor_set_info(processor_set_t pset,
					int flavor,
					host_t *host,
					processor_set_info_t info,
					int *count);

kern_return_t my_processor_set_info(PatchPtr patch,
				    int flavor,
				    host_t *host,
				    processor_set_info_t info,
				    int *count);

  
extern kern_return_t processor_set_statistics(processor_set_t pset,
					      int flavor,
					      processor_set_info_t info,
					      int *count);
  
kern_return_t my_processor_set_statistics(PatchPtr patch,
					  int flavor,
					  processor_set_info_t info,
					  int *count);


extern kern_return_t host_info(host_t host, host_flavor_t flavor,
			       host_info_t info, int *count);

kern_return_t my_host_info(PatchPtr patch, host_flavor_t flavor,
			   host_info_t info, int *count);

extern kern_return_t host_processor_slots(host_t host, processor_slot_t slots,
					  int *count);

kern_return_t my_host_processor_slots(PatchPtr patch, processor_slot_t slots,
				      int *count);

extern kern_return_t processor_start(processor_t processor);
kern_return_t my_processor_start(PatchPtr patch);

extern kern_return_t processor_exit(processor_t processor);
kern_return_t my_processor_exit(PatchPtr patch);

#define kSlowFactor (3)
static long gSecondCPUActive;

kern_return_t MPEnablerStart(struct kmod_info *ki, void *data)
{
  InitPatchIt();
  
  PatchIt(processor_set_info, my_processor_set_info);
  PatchIt(processor_set_statistics, my_processor_set_statistics);
  PatchIt(host_info, my_host_info);
  PatchIt(host_processor_slots, my_host_processor_slots);
  PatchIt(processor_start, my_processor_start);
  PatchIt(processor_exit, my_processor_exit);
  
  SetICTC(1, kSlowFactor);
  
  return KERN_SUCCESS;
}

kern_return_t MPEnablerStop(struct kmod_info *ki, void *data)
{
  UnInitPatchIt();
  
  return KERN_SUCCESS;
}


void SetICTC(long on, long cycles)
{
  long tmp;
  
  tmp = cycles << 1;
  if (on) tmp |= 1;
  
  __asm__ volatile("sync");
  __asm__ volatile("isync");
  __asm__ volatile("mtspr 1019, %0" : : "r" (tmp));
  __asm__ volatile("isync");
}


kern_return_t my_processor_set_info(PatchPtr patch,
				    int flavor,
				    host_t *host,
				    processor_set_info_t info,
				    int *count)
{
  kern_return_t ret;
  processor_set_t pset = (processor_set_t)patch->arg0Save;
  PatchProcPtr patchProc = (PatchProcPtr)(patch->patchStub);
  
  ret = patchProc(pset, flavor, host, info, count);
  
  if (ret == KERN_SUCCESS) {
    if (gSecondCPUActive) {
      if (flavor == PROCESSOR_SET_BASIC_INFO) {
	register processor_set_basic_info_t basic_info;
	
	basic_info = (processor_set_basic_info_t)info;
	
	basic_info->processor_count = 2;
      }
    }
  }
  
  return ret;
}

kern_return_t my_processor_set_statistics(PatchPtr patch,
					  int flavor,
					  processor_set_info_t info,
					  int *count)
{
  kern_return_t ret;
  processor_set_t pset = (processor_set_t)patch->arg0Save;
  PatchProcPtr patchProc = (PatchProcPtr)(patch->patchStub);
  
  ret = patchProc(pset, flavor, info, count);
  
  if (ret == KERN_SUCCESS) {
    if (gSecondCPUActive) {
      if (flavor == PROCESSOR_SET_LOAD_INFO) {
	register processor_set_load_info_t load_info;
	
	load_info = (processor_set_load_info_t)info;
	
	load_info->mach_factor /= 2;
	load_info->load_average /= 2;
      }
    }
  }
  
  return ret;
}

kern_return_t my_host_info(PatchPtr patch, host_flavor_t flavor,
			   host_info_t info, int *count)
{
  kern_return_t ret;
  host_t host = (host_t)patch->arg0Save;
  PatchProcPtr patchProc = (PatchProcPtr)(patch->patchStub);
  
  ret = patchProc(host, flavor, info, count);
  
  if (ret == KERN_SUCCESS) {
    if (flavor == HOST_BASIC_INFO) {
      register host_basic_info_t basic_info;
      
      basic_info = (host_basic_info_t)info;
      
      basic_info->avail_cpus = 2;
    }
  }
  
  return ret;
}

kern_return_t my_host_processor_slots(PatchPtr patch, processor_slot_t slots,
				      int *count)
{
  kern_return_t ret;
  host_t host = (host_t)patch->arg0Save;
  PatchProcPtr patchProc = (PatchProcPtr)(patch->patchStub);
  
  ret = patchProc(host, slots, count);
  
  if (ret == KERN_SUCCESS) {
    if (gSecondCPUActive) {
      slots[1] = 1;
      (*count)++;
    }
  }
  
  return ret;
}

kern_return_t my_processor_start(PatchPtr patch)
{
  gSecondCPUActive = 1;
  
  SetICTC(0, 0);
  
  return KERN_SUCCESS;
}

kern_return_t my_processor_exit(PatchPtr patch)
{
  gSecondCPUActive = 0;
  
  SetICTC(1, kSlowFactor);
  
  return KERN_SUCCESS;
}
