import W
from WASTEconst import *
import PyEdit
import PyFontify
import string
import Qd
import Fm

kStringColor = (0, 0x7fff, 0)
kCommentColor = (0, 0, 0xb000)

class DirtTracker:
	"""This class keeps track of ranges which are dirty (i.e., need to be
	updated or processed), even when some of those may overlap.  It can then
	return a range at a time for processing.
	
	Experimental: it also keeps one open triple-quote range, for tracking
	where you've got a triple-quote in progress."""
	
	def __init__(self):
		self.ranges = []
		self.tripleQ = None		# when there's a """ open, this is [from,to] offsets
	
	def PushRange(self, range):
		"Add a range to our list of ranges that need updating."
		# Combine any overlapping ranges for max. efficiency.
		for r in self.ranges:
			if r[0] < range[1] and range[0] < r[1]:
				r[0] = min(r[0], range[0])
				r[1] = max(r[1], range[1])
				return
		self.ranges.append(list(range))
#		self.Spew()
	
	def PopRange(self):
		"Return a range that needs updating, and remove it from the list."
		if not self.ranges: return None
		out = self.ranges[-1]
		del self.ranges[-1]
#		self.Spew()
		return out
	
	def Adjust(self, adjustLoc=0, adjustment=1):
		"Add adjustment to every element >= adjustLoc."
		for r in self.ranges:
			if r[0] >= adjustLoc: r[0] = r[0] + adjustment
			if r[1] >= adjustLoc: r[1] = r[1] + adjustment
		if self.tripleQ:
			if self.tripleQ[0] >= adjustLoc: self.tripleQ[0] = self.tripleQ[0] + adjustment
			if self.tripleQ[1] >= adjustLoc: self.tripleQ[1] = self.tripleQ[1] + adjustment
			
	
	#def Spew(self):
	#	"Print current ranges to sys.stdout."
	#	print "Dirty:", self.ranges
		

class FontifiedPyEditWidget(W.PyEditor):
	
	def __init__(self, possize, text = "", callback = None, inset = (4, 4),
				fontsettings = ("Python-Sans", 0, 9, (0, 0, 0)),
				tabsettings = (32, 0),
				readonly = 0,
				debugger = None,
				file = ''):
		self.inherited = W.PyEditor
		self.inherited.__init__(self, possize, text, callback, inset, fontsettings, tabsettings, readonly, debugger, file)
		self.dirtTracker = DirtTracker()
		self.maxDirtSize = 40*40		# about 40 lines?
		self.getenvironment = None		# we need this to be assigned by our PyEditor

	def _getflags(self):
		return weDoUndo | weDoDrawOffscreen | weDoUseTempMem | weDoAutoScroll | weDoOutlineHilite
	
	def checkLine(self, offset):
		"""Check the line containing the given offset, and see whether it begins or
		ends a triple-quote.  Update lineNotes accordingly."""
		linerange = self.ted.WEFindLine( offset, kLeadingEdge )
		line = self.get()[linerange[0] : linerange[1]]	# isn't there an easier way?!?
		qpos = string.find(line, '"""')
		if qpos < 0: return		# no triple quotes started or ended here!
		# ok, so it looks like there's a """ on this line...
		

	def markDirty(self, fromWhere, toWhere ):
		"""Mark the given fromWhere,toWhere range as dirty, padding to lines and breaking
		into smaller chunks as necessary."""
		
		# expand to the start and end of the line
		beginning = self.ted.WEFindLine( fromWhere, kLeadingEdge )[0]
		ending = self.ted.WEFindLine( toWhere, kLeadingEdge )[1]
		
		# also check whether we're within an open triple-quote...
		# if we are, expand to fill that range
		if self.dirtTracker.tripleQ and ending >= self.dirtTracker.tripleQ[0] \
				and beginning <= self.dirtTracker.tripleQ[1]:
			beginning = min(beginning, self.dirtTracker.tripleQ[0])
			ending = max(ending, self.dirtTracker.tripleQ[1])
		else:
			# also check the run range -- see if this line's already in a string
			# (and if so, expand to include the whole string)
			runInfo = self.ted.WEGetRunInfo( beginning )
			runRange = runInfo[:2]
			runColor = runInfo[4][3]
			if runColor == kStringColor:
				beginning = runRange[0]
				ending = runRange[1]
		
		if ending - beginning > self.maxDirtSize:
			parts = range(beginning, ending, self.maxDirtSize) + [ending]
			for i in range(0, len(parts)-1):
				self.dirtTracker.PushRange( (parts[i], parts[i+1]) )
			return
		self.dirtTracker.PushRange( (beginning, ending) )
		
	def fontify(self):
		self.fontifypart()
	
	def fontifyselection(self):
		selstart, selend = self.ted.WEGetSelection()
		self.fontifypart(selstart, selend)
	
	def fontifyDirty(self):
		part = self.dirtTracker.PopRange()
		if part:
#			print "Fontifying:", part
			self.fontifypart( part[0], part[1] )
		
	def fontifypart(self, start = 0, end = None):
		W.SetCursor('watch')
		self.ted.WEFeatureFlag(weFOutlineHilite, 0)
		self.ted.WEDeactivate()
		#self.ted.WEFeatureFlag(weFInhibitRecal, 1)
		#self.ted.WEFeatureFlag(weFInhibitRedraw, 1)
		self.ted.WEFeatureFlag(weFAutoScroll, 0)
		self.ted.WEFeatureFlag(weFUndo, 0)
		pytext = string.join(string.split(self.get(), '\r'), '\n')
		if end is None:
			end = len(pytext)
		selstart, selend = self.ted.WEGetSelection()
		self.ted.WESetSelection(start, end)
		self.ted.WESetStyle(weDoFace | weDoColor | weDoFont | weDoSize, 
				(Fm.GetFNum('Profont'), 0, 10, (0, 0, 0)))
		
		tags = PyFontify.fontify(pytext, start, end)
		styles = {
			'string': (weDoColor, (0, 0, 0, kStringColor)),
			'keyword': (weDoFace, (0, 1, 0, (0, 0, 0))),
			'comment': (weDoFace | weDoColor, (0, 2, 0, kCommentColor)),
			'identifier': (weDoColor, (0, 0, 0, (0xbfff, 0, 0)))
		}
		setselection = self.ted.WESetSelection
		setstyle = self.ted.WESetStyle
		for tag, start, end, sublist in tags:
			setselection(start, end)
			mode, style = styles[tag]
			setstyle(mode, style)
		self.ted.WESetSelection(selstart, selend)
		self.SetPort()
		#Qd.EraseRect(self.ted.WEGetViewRect())
		#self.ted.WEFeatureFlag(weFInhibitRedraw, 0)
		#self.ted.WEFeatureFlag(weFInhibitRecal, 0)
		self.ted.WEFeatureFlag(weFAutoScroll, 1)
		self.ted.WEFeatureFlag(weFUndo, 1)
		self.ted.WEActivate()
		self.ted.WEFeatureFlag(weFOutlineHilite, 1)
		#self.ted.WECalText()
		#self.ted.WEUpdate(self._parentwindow.wid.GetWindowPort().visRgn)

	def idle(self):
		"Fontify the dirty parts, then do regular idling."
		self.fontifyDirty()
		self.inherited.idle(self)
		
	def handleTripleQuote(self, offset):
		"Handle a newly entered triple-quote at location 'offset'."
		
	def HandleQuestion(self, offset):
		"Handle a question mark by looking up the preceeding function. \
		Return 1/0 for success/fail."
		# find the preceeding word
		wordRange = self.ted.WEFindWord( offset-1, kTrailingEdge )
		word = self.get()[wordRange[0]:wordRange[1]]	# is there some more efficient way?
		print "Querying:", word

		# now, we need to find the function matching this word...
		try:
			# try looking it up in the code's environment
			globals, file, modname = self.getenvironment()
			f = eval(word, globals)
		except: f = None
		# try looking it up in our own namespace
		try: f = eval(word)
		except: pass
		if not callable(f):
			# try splitting it into parts, and importing the first part
			modname = string.split(word,'.')[0]
			try:
				mod = __import__(modname)
				f = mod.__dict__[string.split(word,'.')[1]]
			except:pass
		
		try:
			d = dir(f)
			if "func_code" in d:
				defoffset = f.func_code.co_argcount - len(f.func_defaults)
				args = []
				for i in range(0, f.func_code.co_argcount):
					args.append( f.func_code.co_varnames[i] )
					if i >= defoffset: args[-1] = args[-1] + "=%s" % f.func_defaults[i-defoffset]
				print "%s(%s)" % (f.func_name, string.join(args,','))
			if f.__doc__:
				print f.__doc__
			else:
				print f
			return 1
		except:
			return 0
		
		
	def key(self, char, event):
		# if it's not a typing key, then just do the inherited stuff
		# LATER: watch for special keys like delete, del, etc.
		if (char != chr(8) and char < ' ' or char > chr(127)):
			#print "char:", ord(char)
			self.inherited.key(self, char, event)
			return
			
		selRange = self.ted.WEGetSelection()

		# check for special case of "?" (for function info)
		if char == '?':
			if self.HandleQuestion(selRange[0]):
				return		# query was successful, so don't enter the ? mark
		
		# for now, let's assume it's a typing key
		effectFrom = self.ted.WEFindLine(selRange[0], kLeadingEdge)
		effectTo = self.ted.WEFindLine(selRange[1], kLeadingEdge)
		fullEffect = (effectFrom[0], effectTo[1])

		# so we're replacing selRange with 0 or 1 chars...
		if char == chr(8): newlen = 0		# delete key -- no new chars
		else: newlen = 1					# anything else -- 1 new char
		# this requires an adjustment
		self.dirtTracker.Adjust( adjustLoc=selRange[0], adjustment=-(selRange[1]-selRange[0])+newlen )
		
		#print "Got key %s affecting %s" % (char, fullEffect)		

		# do the key
		self.inherited.key(self, char, event)

		# check for special case of triple-quotes
		if char == '"':
			quote = ord(char)
			if WEGetChar(selRange[0]-1) == quote and WEGetChar(selRange[0]-2) == quote:
				# we just typed a triple quote!!!
				self.handleTripleQuote(selRange[0]-2)
					
		# now, note what range needs updated
		self.markDirty( effectFrom[0], effectFrom[0] )

	def open(self):
		self.inherited.open(self)
		self.markDirty( 0, self.ted.WEGetTextLength() )
		
	def domenu_clear(self, *args):
		selRange = self.ted.WEGetSelection()
		self.dirtTracker.Adjust( adjustLoc=selRange[0], adjustment=-(selRange[1]-selRange[0]) )
		self.inherited.domenu_clear(self, args)	
		self.markDirty( selRange[0], selRange[0] )

	def domenu_cut(self, *args):
		selRange = self.ted.WEGetSelection()
		self.dirtTracker.Adjust( adjustLoc=selRange[0], adjustment=-(selRange[1]-selRange[0]) )
		self.inherited.domenu_cut(self, args)
		self.markDirty( selRange[0], selRange[0] )
		
	def domenu_paste(self, *args):
		selRange = self.ted.WEGetSelection()
		# adjustment depends on how big the text pasted is!
		# we can get that by doing the paste, then checking the cursor again.
		self.inherited.domenu_paste(self, args)
		newRange = self.ted.WEGetSelection()
		adjustment = newRange[0] - (selRange[1]-selRange[0])
		self.dirtTracker.Adjust( adjustLoc=selRange[0], adjustment=adjustment )
		# then, note that the newly pasted stuff is dirty
		self.markDirty( selRange[0], newRange[0] )
				
	def domenu_comment(self):
		self.inherited.domenu_comment(self)
		selRange = self.ted.WEGetSelection()
		self.markDirty( selRange[0], selRange[1] )

	def domenu_uncomment(self):
		self.inherited.domenu_uncomment(self)
		selRange = self.ted.WEGetSelection()
		self.markDirty( selRange[0], selRange[1] )

class FontifiedPyEditor(PyEdit.Editor):
	
	def setupwidgets(self, text):
		topbarheight = 24
		popfieldwidth = 80
		self.lastlineno = None
		
		# make an editor
		self.editgroup = W.Group((0, topbarheight + 1, 0, 0))
		self.editgroup.editor = FontifiedPyEditWidget((0, 0, -15,-15), text)
		self.editgroup.editor.getenvironment = self.getenvironment
			
		# make the widgets
		self.popfield = PyEdit.ClassFinder((popfieldwidth - 17, -15, 16, 16), [], self.popselectline)
		self.linefield = W.EditText((-1, -15, popfieldwidth - 15, 16), inset = (6, 2))
		self.editgroup._barx = W.Scrollbar((popfieldwidth-2, -15, -14, 16), self.editgroup.editor.hscroll, max = 32767)
		self.editgroup._bary = W.Scrollbar((-15, 14, 16, -14), self.editgroup.editor.vscroll, max = 32767)
		self.editgroup.optionsmenu = W.PopupMenu((-15, -1, 16, 16), ['abc', 'defg'])
		self.hline = W.HorizontalLine((0, topbarheight, 0, 0))
		self.infotext = W.TextBox((175, 4, -4, 14))
		self.runbutton = W.Button((5, 4, 80, 16), "Run all", self.run)
		self.runselbutton = W.Button((90, 4, 80, 16), "Run selection", self.runselection)
		
		# bind some keys
		self.editgroup.editor.bind("cmdr", self.runbutton.push)
		self.editgroup.editor.bind("enter", self.runselbutton.push)
		self.editgroup.editor.bind("cmdj", self.domenu_gotoline)
		self.editgroup.editor.bind("<idle>", self.updateselection)
		
		self.editgroup.editor.bind("cmde", PyEdit.searchengine.setfindstring)
		self.editgroup.editor.bind("cmdf", PyEdit.searchengine.show)
		self.editgroup.editor.bind("cmdg", PyEdit.searchengine.findnext)
		self.editgroup.editor.bind("cmdshiftr", PyEdit.searchengine.replace)
		self.editgroup.editor.bind("cmdt", PyEdit.searchengine.replacefind)
		
		self.linefield.bind("return", self.dolinefield)
		self.linefield.bind("enter", self.dolinefield)
		self.linefield.bind("tab", self.dolinefield)
		
		self.bind('cmd1', self.editgroup.editor.fontify)
		self.bind('cmd2', self.editgroup.editor.fontifyselection)
		self.bind('cmd3', self.editgroup.editor.fontifyDirty)
		
		# intercept clicks
		self.editgroup.editor.bind("<click>", self.clickeditor)
		self.linefield.bind("<click>", self.clicklinefield)


def test():
	import macfs
	fsspec, ok = macfs.StandardGetFile('TEXT')
	if not ok: return FontifiedPyEditor()
	else: return FontifiedPyEditor(fsspec.as_pathname())

e = test()