// KeyEventHandler.cp

#ifndef KeyEventHandler_h
#include "KeyEventHandler.h"
#endif
#ifndef KeyDownEvent_h
#include "KeyDownEvent.h"
#endif
#ifndef AutoKeyEvent_h
#include "AutoKeyEvent.h"
#endif
#ifndef UserState_h
#include "UserState.h"
#endif
#ifndef UserCancelledError_h
#include "UserCancelledError.h"
#endif
#ifndef MenuBar_h
#include "MenuBar.h"
#endif
#ifndef FunctionKeys_h
#include "FunctionKeys.h"
#endif
#ifndef Commander_h
#include "Commander.h"
#endif
#ifndef TabKeys_h
#include "TabKeys.h"
#endif
#ifndef DeleteKeys_h
#include "DeleteKeys.h"
#endif
#ifndef ReturnKeys_h
#include "ReturnKeys.h"
#endif
#ifndef StopKeys_h
#include "StopKeys.h"
#endif
#ifndef EscapeKeys_h
#include "EscapeKeys.h"
#endif
#ifndef Undoing_h
#include "Undoing.h"
#endif
#ifndef Editing_h
#include "Editing.h"
#endif
#ifndef HelpKeys_h
#include "HelpKeys.h"
#endif
#ifndef Scrolling_h
#include "Scrolling.h"
#endif
#ifndef HorizontalArrows_h
#include "HorizontalArrows.h"
#endif
#ifndef VerticalArrows_h
#include "VerticalArrows.h"
#endif
#ifndef BufferOfSize_h
#include "BufferOfSize.h"
#endif
#ifndef EventQueue_h
#include "EventQueue.h"
#endif
#ifndef Typing_h
#include "Typing.h"
#endif

void KeyEventHandler::HandleKeyDownEvent( const KeyDownEvent& event )
  {
	ObscureCursor();
	
	try
	  {
		DispatchKeystroke( event );
	  }
	 catch ( const UserCancelledError& )
	  {
	  }
	
	UserState::The().Announce();
  }

void KeyEventHandler::HandleAutoKeyEvent( const AutoKeyEvent& event )
  {
	ObscureCursor();

	try
	  {
		DispatchKeystroke( event );
	  }
	 catch ( const UserCancelledError& )
	  {
	  }
	
	UserState::The().Announce();
  }

void KeyEventHandler::HandleKeyUpEvent( const KeyUpEvent& )
  {
  }

void KeyEventHandler::DispatchKeystroke( const KeystrokeEvent& event )
  {
	if ( event.Key().Special() )
		DispatchSpecialKey( event );
	 else
		if ( event.Command() )
			DispatchCommandKey( event );
		 else
			DispatchTypingKey( event );
  }

void KeyEventHandler::DispatchCommandKey( const KeystrokeEvent& event )
  {
	MenuBar::The().Key( event );
  }

void KeyEventHandler::DispatchTypingKey( const KeystrokeEvent& event )
  {
	// We'll buffer up as many typing keys as we can get, for
	// two purposes: we'll keep up with fast typists, and we'll
	// have an easier time dealing with multibyte characters,
	// which come as multiple events.
	
	BufferOfSize<256> buffer;
	buffer.Put( event.Key().Character() );

	for ( EventQueue& queue( EventQueue::The() );
			!buffer.Exhausted();
			queue++ )
	  {
		const EventRecord& nextEvent( queue.Next() );
		
		if ( nextEvent.what != keyDown && nextEvent.what != autoKey )
			break;
		
		KeystrokeEvent nextKeystroke( nextEvent );
		
		if ( nextKeystroke.Command() )
			break;
		
		if ( nextKeystroke.Key().Special() )
			break;
		
		buffer.Put( nextKeystroke.Key().Character() );
	  }
	
	if ( !Commander<Typing>::Null() )
		Commander<Typing>()->Type( buffer.Used() );
  }

void KeyEventHandler::DispatchSpecialKey( const KeystrokeEvent& event )
  {
	switch ( event.Key().Code() )
	  {
		case Key::tabKey:
			if ( !Commander<TabKeys>::Null() )
			  {
				if ( event.Shift() )
					Commander<TabKeys>()->TabBackward( event );
			 	else
					Commander<TabKeys>()->TabForward( event );
			  }
			break;
		
		case Key::deleteKey:
			if ( !Commander<DeleteKeys>::Null() )
			  {
				if ( event.Shift() )
					Commander<DeleteKeys>()->DeleteForward( event );
				 else
					Commander<DeleteKeys>()->DeleteBackward( event );
			  }
			break;

		case Key::forwardDeleteKey:
			if ( !Commander<DeleteKeys>::Null() )
			  {
				if ( event.Shift() )
					Commander<DeleteKeys>()->DeleteBackward( event );
				 else
					Commander<DeleteKeys>()->DeleteForward( event );
			  }
			break;

		case Key::returnKey:
			if ( !Commander<ReturnKeys>::Null() )
				Commander<ReturnKeys>()->ReturnKey( event );
			break;
		
		case Key::enterKey:
			if ( !Commander<ReturnKeys>::Null() )
				Commander<ReturnKeys>()->EnterKey( event );
			break;
		
		case Key::stopKey:
			if ( !Commander<StopKeys>::Null() )
				Commander<StopKeys>()->StopKey( event );
			break;

		case Key::escapeKey:
			if ( !Commander<EscapeKeys>::Null() )
				Commander<EscapeKeys>()->EscapeKey( event );
			break;

		case Key::function1:
			if ( !Commander<Undoing>::Null() )
				Commander<Undoing>()->UndoOrRedo();
			break;
		
		case Key::function2:
			if ( !Commander<Editing>::Null() )
				Commander<Editing>()->Cut();
			break;

		case Key::function3:
			if ( !Commander<Editing>::Null() )
				Commander<Editing>()->Copy();
			break;

		case Key::function4:
			if ( !Commander<Editing>::Null() )
				Commander<Editing>()->Paste();
			break;

		case Key::clearKey:
			if ( !Commander<Editing>::Null() )
				Commander<Editing>()->Clear();
			break;

		case Key::function5:
			if ( !Commander<FunctionKeys>::Null() )
				Commander<FunctionKeys>()->FunctionKey( 5, event );
			break;

		case Key::function6:
			if ( !Commander<FunctionKeys>::Null() )
				Commander<FunctionKeys>()->FunctionKey( 6, event );
			break;

		case Key::function7:
			if ( !Commander<FunctionKeys>::Null() )
				Commander<FunctionKeys>()->FunctionKey( 7, event );
			break;

		case Key::function8:
			if ( !Commander<FunctionKeys>::Null() )
				Commander<FunctionKeys>()->FunctionKey( 8, event );
			break;

		case Key::function9:
			if ( !Commander<FunctionKeys>::Null() )
				Commander<FunctionKeys>()->FunctionKey( 9, event );
			break;

		case Key::function10:
			if ( !Commander<FunctionKeys>::Null() )
				Commander<FunctionKeys>()->FunctionKey( 10, event );
			break;

		case Key::function11:
			if ( !Commander<FunctionKeys>::Null() )
				Commander<FunctionKeys>()->FunctionKey( 11, event );
			break;

		case Key::function12:
			if ( !Commander<FunctionKeys>::Null() )
				Commander<FunctionKeys>()->FunctionKey( 12, event );
			break;

		case Key::function13:
			if ( !Commander<FunctionKeys>::Null() )
				Commander<FunctionKeys>()->FunctionKey( 13, event );
			break;

		case Key::function14:
			if ( !Commander<FunctionKeys>::Null() )
				Commander<FunctionKeys>()->FunctionKey( 14, event );
			break;

		case Key::function15:
			if ( !Commander<FunctionKeys>::Null() )
				Commander<FunctionKeys>()->FunctionKey( 15, event );
			break;

		case Key::helpKey:
			if ( !Commander<HelpKeys>::Null() )
				Commander<HelpKeys>()->HelpKey( event );
			break;

		case Key::homeKey:
			if ( !Commander<Scrolling>::Null() )
				Commander<Scrolling>()->ScrollToHome();
			break;

		case Key::endKey:
			if ( !Commander<Scrolling>::Null() )
				Commander<Scrolling>()->ScrollToEnd();
			break;

		case Key::pageUpKey:
			if ( !Commander<Scrolling>::Null() )
				Commander<Scrolling>()->ScrollUpOnePage();
			break;

		case Key::pageDownKey:
			if ( !Commander<Scrolling>::Null() )
				Commander<Scrolling>()->ScrollDownOnePage();
			break;

		case Key::leftArrow:
			if ( !Commander<HorizontalArrows>::Null() )
				Commander<HorizontalArrows>()->LeftArrow( event );
			break;

		case Key::rightArrow:
			if ( !Commander<HorizontalArrows>::Null() )
				Commander<HorizontalArrows>()->RightArrow( event );
			break;

		case Key::upArrow:
			if ( !Commander<VerticalArrows>::Null() )
				Commander<VerticalArrows>()->UpArrow( event );
			break;

		case Key::downArrow:
			if ( !Commander<VerticalArrows>::Null() )
				Commander<VerticalArrows>()->DownArrow( event );
			break;

		case Key::power:
			Assert( 0 );
			break;
		
		default:
			Assert( 0 );
			break;
	  }
  }
