// ControlObject.cp

#ifndef ControlObject_h
#include "ControlObject.h"
#endif
#ifndef WindowObject_h
#include "WindowObject.h"
#endif
#ifndef MemoryFullError_h
#include "MemoryFullError.h"
#endif

// This is a cheap compile-time assertion that
// the sizes are equal:
	static const void *shouldBeNull
		= sizeof(ControlHandle) - sizeof(ControlObject);

ControlObject::ControlObject( WindowObject& window,
										ControlDefinition definition,
										const Rect& bounds,
										ConstPString title,
										bool visible,
										int16 minimum,
										int16 value,
										int16 maximum,
										uint32 refCon )
  {
	control = NewControl( &window.Port(),
								 &bounds,
								 title,
								 visible,
								 value,
								 minimum,
								 maximum,
								 definition.Number(),
								 refCon );
	
	Assert( control != 0 );
	if ( control == 0 )
		throw MemoryFullError();
	Assert( *control != 0 );
  }

ControlObject::~ControlObject()
  {
	SetControlVisibility( control, false, false );
	DisposeControl( control );
  }

void ControlObject::SetValue( int16 v )
  {
	Assert( Minimum() <= v );
	Assert( v <= Maximum() );
	if ( v != Value() )
		SetControlValue( control, v );
  }

void ControlObject::SetMinimum( int16 m )
  {
	Assert( m <= Value() );
	if ( m != Minimum() )
		SetControlMinimum( control, m );
  }

void ControlObject::SetMaximum( int16 m )
  {
	Assert( m >= Value() );
	if ( m != Maximum() )
		SetControlMaximum( control, m );
  }

void ControlObject::SetRangeAndValue( int16 minimum,
												  int16 value,
												  int16 maximum )
  {
	Assert( minimum <= value );
	Assert( value <= maximum );
	
	bool changeMinimum = ( minimum != Minimum() );
	bool changeValue =   ( value != Value() );
	bool changeMaximum = ( maximum != Maximum() );
	
	if ( !changeValue && !changeMaximum )
	  {
		if ( changeMinimum )
			SetControlMinimum( control, minimum );
		return;
	  }
	
	if ( !changeMinimum && !changeMaximum )
	  {
		SetControlValue( control, value );
		return;
	  }
	
	if ( !changeMinimum && !changeValue )
	  {
		SetControlMaximum( control, maximum );
		return;
	  }
	
	bool wasVisible = Visible();
	
	if ( wasVisible )
		SetControlVisibility( control, false, false );

	if ( changeMinimum )
		SetControlMinimum( control, minimum );
	
	if ( changeMaximum )
		SetControlMaximum( control, maximum );
	
	if ( changeValue )
		SetControlValue( control, value );

	if ( wasVisible )
	  {
		SetControlVisibility( control, true, false );
		Draw();
	  }
  }

void ControlObject::SetTitle( ConstPString s )
  {
	String255 title;
	GetTitle( title );
	if ( title != s )
		SetControlTitle( control, s );
  }

void ControlObject::SetBounds( const Rectangle& newBounds )
  {
	Rectangle oldBounds( Bounds() );
	
	if ( oldBounds.TopLeft() != newBounds.TopLeft() )
		MoveTo( newBounds.TopLeft() );

	if ( oldBounds.Size() != newBounds.Size() )
		SetSize( newBounds.Size() );
  }

PointObject ControlObject::RecommendedSize() const
  {
	Rectangle best( 0, 0, 0, 0 );
	int16 baseLine = 0;
	
	ThrowOSError( GetBestControlRect( control, &best, &baseLine ) );
	
	return best.Size();
  }

int16 ControlObject::Track( PointObject start, ControlActionUPP action )
  {
	return TrackControl( control, start, action );
  }
