// ControlView.cp

#ifndef ControlView_h
#include "ControlView.h"
#endif
#ifndef Window_h
#include "Window.h"
#endif
#ifndef ListLoop_h
#include "ListLoop.h"
#endif
#ifndef MouseDownEvent_h
#include "MouseDownEvent.h"
#endif
#ifndef Canvas_h
#include "Canvas.h"
#endif
#ifndef ControlDrawingContext_h
#include "ControlDrawingContext.h"
#endif
#ifndef ControlPart_h
#include "ControlPart.h"
#endif

const Rect ControlView::defaultBounds = { 0, 0, 100, 200 };

ControlView::ControlView( Window& theWindow,
								  ControlDefinition definition,
								  ConstPString title,
								  int16 minimum,
								  int16 value,
								  int16 maximum,
								  Rectangle theViewInset )
  : Activator( theWindow.Focus() ),
	 control( theWindow.WindowObject(),
				 definition,
				 defaultBounds,
				 title,
				 false,
				 minimum,
				 value,
				 maximum,
				 0 ),
	 viewInset( theViewInset ),
	 window( theWindow ),
	 visibility( 0 )
  {
	Assert( minimum <= value );
	Assert( value <= maximum );
  }

ControlView::ControlView( Window& theWindow,
								  ControlDefinition definition,
								  ConstPString title,
								  int16 magicMinimum,
								  int16 magicValue,
								  int16 magicMaximum,
								  int16 minimum,
								  int16 value,
								  int16 maximum,
								  Rectangle theViewInset )
  : Activator( theWindow.Focus() ),
	 control( theWindow.WindowObject(),
				 definition,
				 defaultBounds,
				 title,
				 false,
				 magicMinimum,
				 magicValue,
				 magicMaximum,
				 0 ),
	 viewInset( theViewInset ),
	 window( theWindow ),
	 visibility( 0 )
  {
	Assert( minimum <= value );
	Assert( value <= maximum );	
	control.SetRangeAndValue( minimum, value, maximum );
  }

ControlView::~ControlView()
  {
  }

void ControlView::PrepareToDraw( const Canvas& canvas )
  {
	Assert( visibility < maxuint32 );
	if ( visibility == 0 )
	  {
		PointObject position = canvas.ToScreen( viewInset.TopLeft() );
		if ( control.Bounds().TopLeft() != position )
			control.MoveTo( position );

		Assert( !control.Visible() );
		control.Show();
	  }
	visibility++;
  }

void ControlView::StopDrawing()
  {
	Assert( visibility > 0 );
	visibility--;
	if ( visibility == 0 )
	  {
		Assert( control.Visible() );
		control.Hide();
	  }
  }

void ControlView::Draw( const Canvas& canvas ) const
  {
	ControlView *mutableThis = const_cast<ControlView *>( this );
	mutableThis->PrepareToDraw( canvas );
	control.ControlObject::Draw();
	mutableThis->StopDrawing();
  }

ControlPart *ControlView::FindPart( int16 code ) const
  {
	if ( code == 0 )
		return 0;

	for ( ListLoop<ControlPart> part( parts ); part.Unfinished(); part++ )
		if ( part->Code() == code )
			return &*part;

	return 0;
  }

void ControlView::Click( const Canvas& clicked,
								 const MouseDownEvent& event )
  {
	PrepareToDraw( clicked );
	int16 partNumber = control.Test( event.LocalPoint() );
	StopDrawing();
	
	ControlPart *part = FindPart( partNumber );
	
	if ( part == 0 )
		return;
	
	part->Click( clicked, event );
  }

PointObject ControlView::RecommendedSize() const
  {
	return control.RecommendedSize() - viewInset.Size();
  }

void ControlView::SetValue( int16 v )
  {
	if ( v != control.Value() )
	  {
		ControlDrawingContext context( *this );
		control.SetValue( v );
	  }
  }

void ControlView::SetMinimum( int16 m )
  {
	if ( m != control.Minimum() )
	  {
		ControlDrawingContext context( *this );
		control.SetMinimum( m );
	  }
  }

void ControlView::SetMaximum( int16 m )
  {
	if ( m != control.Maximum() )
	  {
		ControlDrawingContext context( *this );
		control.SetMaximum( m );
	  }
  }

void ControlView::SetRangeAndValue( int16 minimum,
												int16 value,
												int16 maximum )
  {
	if ( minimum != control.Minimum()
	     || maximum != control.Maximum()
	     || value != control.Value() )
	  {
		ControlDrawingContext context( *this );
		control.SetRangeAndValue( minimum, value, maximum );
	  }
  }

void ControlView::SetTitle( ConstPString s )
  {
	String255 title;
	control.GetTitle( title );
	if ( title != s )
	  {
		ControlDrawingContext context( *this );
		control.SetTitle( s );
	  }
  }

void ControlView::Enable()
  {
	ControlDrawingContext context( *this );
	control.Enable();
  }

void ControlView::Disable()
  {
	ControlDrawingContext context( *this );
	control.Disable();
  }

void ControlView::PaneSizeChanged()
  {
	Point32 size = Point32( PaneSize() ) + viewInset.Size();
	control.SetSize( size.NearestPoint16() );
	Invalidate();
  }

void ControlView::Activate()
  {
	ControlDrawingContext context( *this );
	control.Activate();
  }

void ControlView::Deactivate()
  {
	ControlDrawingContext context( *this );
	control.Deactivate();
  }
