// URectangle32.cp

#ifndef URectangle32_h
#include "URectangle32.h"
#endif
#ifndef MinMax_h
#include "MinMax.h"
#endif
#ifndef Overflow_h
#include "Overflow.h"
#endif

const URectangle32 URectangle32::big( 0, 0, maxuint32, maxuint32 );
const URectangle32 URectangle32::zero( 0, 0, 0, 0 );

URectangle32::URectangle32( UPoint32 p1, UPoint32 p2 )
  {
	if ( p1.h <= p2.h )
	  {
		left = p1.h;
		right = p2.h;
	  }
	 else
	  {
		left = p2.h;
		right = p1.h;
	  }
	
	if ( p1.v <= p2.v )
	  {
		top = p1.v;
		bottom = p2.v;
	  }
	 else
	  {
		top = p2.v;
		bottom = p1.v;
	  }
  }

bool URectangle32::Contains( UPoint32 p ) const
  {
	return left <= p.h && p.h < right
		 && top <= p.v && p.v < bottom;
  }

bool URectangle32::StrictlyContains( UPoint32 p ) const
  {
	return left < p.h && p.h < right
		 && top < p.v && p.v < bottom;
  }

bool URectangle32::WeaklyContains( UPoint32 p ) const
  {
	return left <= p.h && p.h <= right
		 && top <= p.v && p.v <= bottom;
  }

bool URectangle32::operator==( URectangle32 r ) const
  {
	return left == r.left && right == r.right 
		 && top == r.top && bottom == r.bottom;
  }

bool URectangle32::operator<=( URectangle32 r ) const
  {
	return left >= r.left && right <= r.right 
		 && top >= r.top && bottom <= r.bottom;
  }

bool URectangle32::operator>=( URectangle32 r ) const
  {
	return left <= r.left && right >= r.right 
		 && top <= r.top && bottom >= r.bottom;
  }

void URectangle32::operator&=( URectangle32 r )
  {
	left = Max( left, r.left );
	top = Max( top, r.top );
	right = Min( right, r.right );
	bottom = Min( bottom, r.bottom );
	
	if ( IsEmpty() )
		*this = URectangle32( left, top, left, top );
  }

void URectangle32::operator|=( URectangle32 r )
  {
	left = Min( left, r.left );
	top = Min( top, r.top );
	right = Max( right, r.right );
	bottom = Max( bottom, r.bottom );
  }

const URectangle32 URectangle32::operator&( URectangle32 r ) const
  {
	URectangle32 result( *this );
	result &= r;
	return result;
  }

const URectangle32 URectangle32::operator|( URectangle32 r ) const
  {
	URectangle32 result( *this );
	result |= r;
	return result;
  }

bool URectangle32::Intersects( URectangle32 r ) const
  {
	if ( left >= r.right || r.left >= right )
		return false;
	
	if ( top >= r.bottom || r.top >= bottom )
		return false;
	
	return true;
  }

bool URectangle32::Touches( URectangle32 r ) const
  {
	if ( left > r.right || r.left > right )
		return false;
	
	if ( top > r.bottom || r.top > bottom )
		return false;
	
	return true;
  }

uint32 URectangle32::SharedSides( URectangle32 r ) const
  {
	return ( ( left   == r.left   ) ? 1 : 0 )
		  + ( ( top    == r.top    ) ? 1 : 0 )
		  + ( ( right  == r.right  ) ? 1 : 0 )
		  + ( ( bottom == r.bottom ) ? 1 : 0 );
  }

void URectangle32::operator+=( URectangle32 r )
  {
	Assert( SharedSides( r ) == 3 );
	Assert( !Intersects( r ) );
	*this |= r;
  }

void URectangle32::operator-=( URectangle32 r )
  {
	Assert( SharedSides( r ) == 3 );
	Assert( *this > r );
	if ( left < r.left )
		right = r.left;
	if ( top < r.top )
		bottom = r.top;
	if ( right > r.right )
		left = r.right;
	if ( bottom > r.bottom )
		top = r.bottom;
  }

const URectangle32 URectangle32::operator+( URectangle32 r ) const
  {
	URectangle32 result( *this );
	result += r;
	return result;
  }

const URectangle32 URectangle32::operator-( URectangle32 r ) const
  {
	URectangle32 result( *this );
	result -= r;
	return result;
  }

void URectangle32::Inset( uint32 amount )
  {
	Assert( CanAdd( top, amount ) );
	Assert( CanAdd( left, amount ) );
	Assert( CanSubtract( bottom, amount ) );
	Assert( CanSubtract( right, amount ) );

	Assert( top + amount <= bottom - amount );
	Assert( left + amount <= right - amount );

	top += amount;
	left += amount;
	bottom -= amount;
	right -= amount;
  }

void URectangle32::Inset( UPoint32 amount )
  {
	Assert( CanAdd( top, amount.v ) );
	Assert( CanAdd( left, amount.h ) );
	Assert( CanSubtract( bottom, amount.v ) );
	Assert( CanSubtract( right, amount.h ) );

	Assert( top + amount.v <= bottom - amount.v );
	Assert( left + amount.h <= right - amount.h );

	top += amount.v;
	left += amount.h;
	bottom -= amount.v;
	right -= amount.h;
  }

void URectangle32::Inset( URectangle32 amount )
  {
	Assert( CanSubtract( top, amount.top ) );
	Assert( CanSubtract( left, amount.left ) );
	Assert( CanSubtract( bottom, amount.bottom ) );
	Assert( CanSubtract( right, amount.right ) );

	Assert( top - amount.top <= bottom - amount.bottom );
	Assert( left - amount.left <= right - amount.right );

	top -= amount.top;
	left -= amount.left;
	bottom -= amount.bottom;
	right -= amount.right;
  }

void URectangle32::Outset( uint32 amount )
  {
	Assert( CanSubtract( top, amount ) );
	Assert( CanSubtract( left, amount ) );
	Assert( CanAdd( bottom, amount ) );
	Assert( CanAdd( right, amount ) );

	Assert( top - amount <= bottom + amount );
	Assert( left - amount <= right + amount );

	top -= amount;
	left -= amount;
	bottom += amount;
	right += amount;
  }

void URectangle32::Outset( UPoint32 amount )
  {
	Assert( CanSubtract( top, amount.v ) );
	Assert( CanSubtract( left, amount.h ) );
	Assert( CanAdd( bottom, amount.v ) );
	Assert( CanAdd( right, amount.h ) );

	Assert( top - amount.v <= bottom + amount.v );
	Assert( left - amount.h <= right + amount.h );

	top -= amount.v;
	left -= amount.h;
	bottom += amount.v;
	right += amount.h;
  }

void URectangle32::Outset( URectangle32 amount )
  {
	Assert( CanAdd( top, amount.top ) );
	Assert( CanAdd( left, amount.left ) );
	Assert( CanAdd( bottom, amount.bottom ) );
	Assert( CanAdd( right, amount.right ) );

	Assert( top + amount.top <= bottom + amount.bottom );
	Assert( left + amount.left <= right + amount.right );

	top += amount.top;
	left += amount.left;
	bottom += amount.bottom;
	right += amount.right;
  }

void URectangle32::operator+=( UPoint32 vector )
  {
	Assert( CanAdd( top, vector.v ) );
	Assert( CanAdd( left, vector.h ) );
	Assert( CanAdd( bottom, vector.v ) );
	Assert( CanAdd( right, vector.h ) );
	top += vector.v;
	left += vector.h;
	bottom += vector.v;
	right += vector.h;
  }

void URectangle32::operator-=( UPoint32 vector )
  {
	Assert( CanSubtract( top, vector.v ) );
	Assert( CanSubtract( left, vector.h ) );
	Assert( CanSubtract( bottom, vector.v ) );
	Assert( CanSubtract( right, vector.h ) );
	top -= vector.v;
	left -= vector.h;
	bottom -= vector.v;
	right -= vector.h;
  }

const URectangle32 URectangle32::operator+( UPoint32 vector ) const
  {
	URectangle32 result( *this );
	result += vector;
	return result;
  }

const URectangle32 URectangle32::operator-( UPoint32 vector ) const
  {
	URectangle32 result( *this );
	result -= vector;
	return result;
  }

const UPoint32 URectangle32::Size() const
  {
	return UPoint32( Width(), Height() );
  }
