// FileReadingPath.cp

#ifndef FileReadingPath_h
#include "FileReadingPath.h"
#endif
#ifndef FileLocation_h
#include "FileLocation.h"
#endif
#ifndef FileNotFoundError_h
#include "FileNotFoundError.h"
#endif
#ifndef DirectoryNotFoundError_h
#include "DirectoryNotFoundError.h"
#endif
#ifndef DiskFullError_h
#include "DiskFullError.h"
#endif
#ifndef FileLockError_h
#include "FileLockError.h"
#endif
#ifndef FileBusyError_h
#include "FileBusyError.h"
#endif
#ifndef FilePermissionError_h
#include "FilePermissionError.h"
#endif
#ifndef HardwareVolumeLockError_h
#include "HardwareVolumeLockError.h"
#endif
#ifndef SoftwareVolumeLockError_h
#include "SoftwareVolumeLockError.h"
#endif
#ifndef BadFileNameError_h
#include "BadFileNameError.h"
#endif

FileReadingPath::FileReadingPath()
  : isOpen( false )
  {
  }

FileReadingPath::FileReadingPath( const FileLocation& location )
  : isOpen( false )
  {
	OpenDataFork( location, FilePermission::ReadOnly() );
  }

FileReadingPath::FileReadingPath( const FileLocation& location, DataFork )
  : isOpen( false )
  {
	OpenDataFork( location, FilePermission::ReadOnly() );
  }

FileReadingPath::FileReadingPath( const FileLocation& location, ResourceFork )
  : isOpen( false )
  {
	OpenResourceFork( location, FilePermission::ReadOnly() );
  }

FileReadingPath::~FileReadingPath()
  {
	// Assert( !IsOpen() );
	// -- No we don't!  There shouldn't be any evil things
	//    to go wrong when closing read-only paths.
	
	if ( IsOpen() )
		try
		  {
			Close();
		  }
		 catch ( OSError& error )
		  {
			error.Debug();
			error.Fix();
		  }
		 catch ( ... )
		  {
			Assert( 0 );
		  }
  }

void FileReadingPath::OpenDataFork( const FileLocation& location,
												FilePermission permission )
  {
	Assert( !IsOpen() );
	Assert( location.NameIsValid() );
	ThrowError( FSpOpenDF( &location, permission.Value(), &refNum ) );
	isOpen = true;
  }

void FileReadingPath::OpenResourceFork( const FileLocation& location,
													 FilePermission permission )
  {
	Assert( !IsOpen() );
	Assert( location.NameIsValid() );
	ThrowError( FSpOpenRF( &location, permission.Value(), &refNum ) );
	isOpen = true;
  }

void FileReadingPath::Close()
  {
	Assert( IsOpen() );
	isOpen = false;
	ThrowError( FSClose( refNum ) );
  }

uint32 FileReadingPath::Length() const
  {
	Assert( IsOpen() );
	int32 length;
	ThrowError( GetEOF( refNum, &length ) );
	return length;
  }

void FileReadingPath::ThrowError( OSErr error )
  {
	if ( error == noErr )
		return;
	
	switch ( error )
	  {
		case fnfErr:				throw FileNotFoundError( error );
		case dirNFErr:				throw DirectoryNotFoundError( error );
		case dskFulErr:			throw DiskFullError( error );
		case permErr:				throw FileLockError( error );
		case opWrErr:				throw FileBusyError( error );
		case afpAccessDenied:	throw FilePermissionError( error );
		
		case wPrErr:				throw HardwareVolumeLockError( error );
		case vLckdErr:				throw SoftwareVolumeLockError( error );
		case fLckdErr:				throw FileLockError( error );
		case bdNamErr:				throw BadFileNameError( error );		
	  }
	
	throw FileError( error );
  }
