// Scrolling.cp

#ifndef Scrolling_h
#include "Scrolling.h"
#endif

UPoint32 Scrolling::ScrollingQuantum() const
  {
	return UPoint32( 1, 1 );
  }

Range32 Scrolling::LineContaining( int32 pixel ) const
  {
	int32 quantum = ScrollingQuantum().v;
	pixel -= pixel % quantum;
	return Range32( pixel, pixel+quantum );
  }

Range32 Scrolling::ColumnContaining( int32 pixel ) const
  {
	int32 quantum = ScrollingQuantum().h;
	pixel -= pixel % quantum;
	return Range32( pixel, pixel+quantum );
  }

void Scrolling::ScrollToHome()
  {
	ScrollTo( Point32::zero );
  }

void Scrolling::ScrollToEnd()
  {
	Point32 panoramaSize = PanoramaSize();
	Point32 revealedSize = RevealedArea().Size();
	
	Point32 target = panoramaSize;
	target.BeBelowRightOf( revealedSize );
	target -= revealedSize;
	
	if ( target.h > 0 )
	  {
		Range32 column = ColumnContaining( target.h );
		if ( column.End() < panoramaSize.h )
			target.h = column.End();
	  }
	
	if ( target.v > 0 )
	  {
		Range32 line = LineContaining( target.v );
		if ( line.End() < panoramaSize.v )
			target.v = line.End();
	  }
	
	ScrollTo( target );	
  }

void Scrolling::ScrollTopTo( int32 newTop )
  {
	ScrollTo( RevealedArea().left, newTop );
  }

void Scrolling::ScrollLeftSideTo( int32 newLeft )
  {
	ScrollTo( newLeft, RevealedArea().top );
  }

void Scrolling::ScrollUp( int32 amount )
  {
	Rectangle32 revealed( RevealedArea() );
	Assert( amount <= revealed.top );
	ScrollTo( revealed.left, revealed.top - amount );
  }

void Scrolling::ScrollDown( int32 amount )
  {
	Rectangle32 revealed( RevealedArea() );
	Assert( amount <= maxint32 - revealed.bottom );
	ScrollTo( revealed.left, revealed.top + amount );
  }

void Scrolling::ScrollLeft( int32 amount )
  {
	Rectangle32 revealed( RevealedArea() );
	Assert( amount <= revealed.left );
	ScrollTo( revealed.left - amount, revealed.top );
  }

void Scrolling::ScrollRight( int32 amount )
  {
	Rectangle32 revealed( RevealedArea() );
	Assert( amount <= maxint32 - revealed.right );
	ScrollTo( revealed.left + amount, revealed.top );
  }

Range32 Scrolling::ReasonableLineSize() const
  {
	int32 quantum = ScrollingQuantum().v;
	
	int32 minimum = quantum + 10;
	minimum -= minimum % quantum;
	Assert( minimum >= quantum );
	 
	int32 maximum = RevealedArea().Height() / 5;
	maximum -= maximum % quantum;
	if ( maximum == 0 )
		maximum = quantum;
	
	return Range32( minimum, maximum );
  }

Range32 Scrolling::ReasonableColumnSize() const
  {
	int32 quantum = ScrollingQuantum().h;
	
	int32 minimum = quantum + 10;
	minimum -= minimum % quantum;
	Assert( minimum >= quantum );
	 
	int32 maximum = RevealedArea().Width() / 5;
	maximum -= maximum % quantum;
	if ( maximum == 0 )
		maximum = quantum;
	
	return Range32( minimum, maximum );
  }
	
void Scrolling::ScrollUpOneLine()
  {
	Rectangle32 revealed( RevealedArea() );
	
	if ( revealed.top == 0 )
		return;

	Range32 reasonable = ReasonableLineSize();
	
	Range32 line = LineContaining( revealed.top - reasonable.Start() );

	if ( revealed.top - line.Start() > reasonable.End() )
		line.SetStart( revealed.top - reasonable.End() );

	ScrollTo( revealed.left, line.Start() );
  }

void Scrolling::ScrollUpOnePage()
  {
	Rectangle32 revealed( RevealedArea() );
	
	if ( revealed.top == 0 )
		return;

	Range32 reasonable = ReasonableLineSize();
	
	Range32 line = LineContaining( revealed.top + reasonable.Start() - 1 );

	if ( line.Start() < revealed.top && line.End() < PanoramaSize().v )
		line = LineContaining( line.End() );

	if ( line.End() - revealed.top > reasonable.End() )
		line.SetEnd( revealed.top + reasonable.End() );
	
	if ( line.End() <= revealed.Height() )
	  {
		ScrollTo( revealed.left, 0 );
		return;
	  }
	
	int32 newTop = line.End() - revealed.Height();
	Range32 newTopLine = LineContaining( newTop );
	
	if ( newTopLine.Start() < newTop
		  && newTopLine.End() + revealed.Height() <= revealed.top + reasonable.End() )
		newTop = newTopLine.End();
	
	ScrollTo( revealed.left, newTop );
  }

void Scrolling::ScrollDownOneLine()
  {
	Rectangle32 revealed( RevealedArea() );
	
	if ( revealed.bottom >= PanoramaSize().v )
		return;

	Range32 reasonable = ReasonableLineSize();
	
	Range32 line = LineContaining( revealed.top + reasonable.Start() - 1 );

	if ( line.End() - revealed.top > reasonable.End() )
		line.SetEnd( revealed.top + reasonable.End() );

	ScrollTo( revealed.left, line.End() );
  }

void Scrolling::ScrollDownOnePage()
  {
	Rectangle32 revealed( RevealedArea() );
	
	if ( revealed.bottom >= PanoramaSize().v )
		return;

	Range32 reasonable = ReasonableLineSize();
	
	Range32 line = LineContaining( revealed.bottom - reasonable.Start() );

	if ( line.End() > revealed.bottom && line.Start() > 0 )
		line = LineContaining( line.Start() - 1 );

	if ( revealed.bottom - line.Start() > reasonable.End() )
		line.SetStart( revealed.bottom - reasonable.End() );
	
	ScrollTo( revealed.left, line.Start() );
  }

void Scrolling::ScrollLeftOneLine()
  {
	Rectangle32 revealed( RevealedArea() );
	
	if ( revealed.left == 0 )
		return;

	Range32 reasonable = ReasonableColumnSize();
	
	Range32 column = ColumnContaining( revealed.left - reasonable.Start() );

	if ( revealed.left - column.Start() > reasonable.End() )
		column.SetStart( revealed.left - reasonable.End() );

	ScrollTo( column.Start(), revealed.top );
  }

void Scrolling::ScrollLeftOnePage()
  {
	Rectangle32 revealed( RevealedArea() );
	
	if ( revealed.left == 0 )
		return;

	Range32 reasonable = ReasonableColumnSize();
	
	Range32 column = ColumnContaining( revealed.left + reasonable.Start() - 1 );

	if ( column.Start() < revealed.left && column.End() < PanoramaSize().h )
		column = ColumnContaining( column.End() );

	if ( column.End() - revealed.left > reasonable.End() )
		column.SetEnd( revealed.left + reasonable.End() );
	
	if ( column.End() <= revealed.Height() )
	  {
		ScrollTo( 0, revealed.top );
		return;
	  }
	
	int32 newLeft = column.End() - revealed.Width();
	Range32 newLeftColumn = ColumnContaining( newLeft );
	
	if ( newLeftColumn.Start() < newLeft
		  && newLeftColumn.End() + revealed.Width() <= revealed.left + reasonable.End() )
		newLeft = newLeftColumn.End();
	
	ScrollTo( newLeft, revealed.top );
  }

void Scrolling::ScrollRightOneLine()
  {
	Rectangle32 revealed( RevealedArea() );
	
	if ( revealed.right >= PanoramaSize().h )
		return;

	Range32 reasonable = ReasonableColumnSize();
	
	Range32 column = ColumnContaining( revealed.left + reasonable.Start() - 1 );

	if ( column.End() - revealed.left > reasonable.End() )
		column.SetEnd( revealed.left + reasonable.End() );

	ScrollTo( column.End(), revealed.top );
  }

void Scrolling::ScrollRightOnePage()
  {
	Rectangle32 revealed( RevealedArea() );
	
	if ( revealed.right >= PanoramaSize().h )
		return;

	Range32 reasonable = ReasonableColumnSize();
	
	Range32 column = ColumnContaining( revealed.right - reasonable.Start() );

	if ( column.End() > revealed.right && column.Start() > 0 )
		column = ColumnContaining( column.Start() - 1 );

	if ( revealed.right - column.Start() > reasonable.End() )
		column.SetStart( revealed.right - reasonable.End() );
	
	ScrollTo( column.Start(), revealed.top );
  }

int32 Scrolling::BestHorizontalPosition( Range32 possible ) const
  {
	Assert( possible.Start() < PanoramaSize().h );
	
	if ( possible.IsEmpty() )
		return possible.Start();
	
	int32 middle = possible.Start() + possible.Length() / 2;
	
	int32 quantum = ScrollingQuantum().h;
	if ( quantum > 1 )
	  {
		middle += quantum / 2;
		middle -= middle % quantum;
	  }
	
	Assert( possible.WeaklyContains( middle ) );

	Range32 column( ColumnContaining( middle ) );
	
	if ( middle - column.Start() <= column.End() - middle )
	  {
		if ( column.Start() >= possible.Start() )
			return column.Start();
	  }
	 else
		if ( column.End() <= possible.End() )
			return column.End();

	return middle;
  }

int32 Scrolling::BestVerticalPosition( Range32 possible ) const
  {
	Assert( possible.Start() < PanoramaSize().v );
	
	if ( possible.IsEmpty() )
		return possible.Start();
	
	int32 middle = possible.Start() + possible.Length() / 2;
	
	int32 quantum = ScrollingQuantum().v;
	if ( quantum > 1 )
	  {
		middle += quantum / 2;
		middle -= middle % quantum;
	  }
	
	Assert( possible.WeaklyContains( middle ) );
	
	Range32 line( LineContaining( middle ) );
	
	if ( middle - line.Start() <= line.End() - middle )
	  {
		if ( line.Start() >= possible.Start() )
			return line.Start();
	  }
	 else
		if ( line.End() <= possible.End() )
			return line.End();

	return middle;
  }
