// ConstArrayOf.h

#ifndef ConstArrayOf_h
#define ConstArrayOf_h

#ifndef Integers_h
#include "Integers.h"
#endif
#ifndef DebugMessage_h
#include "DebugMessage.h"
#endif
#ifndef Assert_h
#include "Assert.h"
#endif
#ifndef Range_h
#include "Range.h"
#endif

// The includes above have been flattened so that weak compilers
// can cope better.

template < class Element >
class ConstArrayOf
  {
	public:
		typedef ConstArrayOf< Element > ConstArrayType;
		typedef int32 (*Comparison)( const Element&, const Element& );

	private:
		const Element *start;
		uint32 length;
	
	public:
		ConstArrayOf()
		  : start( 0 ),
		    length( 0 )
		  {}
		
		ConstArrayOf( const Element *theStart, uint32 theLength )
		  : start( theStart ),
			 length( theLength )
		  { Assert( theStart != 0 || length == 0 ); }

		const Element *Start() const						{ return start; }
		const Element *End() const							{ return start + length; }

		uint32 Length() const								{ return length; }
		const URange32 Range() const						{ return URange32( 0, Length() ); }
				
		uint32 BoundedLength( uint32 bound ) const
			{ return ( length <= bound ) ? length : bound; }

		void LimitLength( uint32 limit )					{ length = BoundedLength( limit ); }
		
		bool IsEmpty() const									{ return length == 0; }
		bool Null() const										{ return start == 0; }
		
		const Element& operator[]( uint32 i ) const	{ Assert( i < length ); return start[i]; }
		
		inline const ConstArrayType Head( uint32 position ) const;
		inline const ConstArrayType Tail( uint32 position ) const;
		inline const ConstArrayType Middle( URange32 ) const;
		
		void Truncate( uint32 position );				// *this = Head( *this, position );
		inline void Shorten( uint32 amount );			// *this = Tail( *this, amount )

		void Append( ConstArrayType tail );
		void Prepend( ConstArrayType head );
		
		void operator+=( ConstArrayType tail )			{ Append( tail ); }
		const ConstArrayType operator+( ConstArrayType tail ) const;
		
		bool Precedes( ConstArrayType r ) const		{ return End() == r.start; }
		bool Follows( ConstArrayType r ) const			{ return start == r.End(); }
		
		bool Contains( const Element *e ) const		{ return start <= e && e < End(); }
		
		bool IsHeadOf( ConstArrayType r ) const;
		bool IsTailOf( ConstArrayType r ) const;
		
		bool IsSortedBy( Comparison ) const;
		
		// These operators express the containment ordering
		bool operator==( ConstArrayType r ) const		{ return start == r.start && length == r.length; }
		bool operator!=( ConstArrayType r ) const		{ return start != r.start || length != r.length; }
		bool operator<=( ConstArrayType r ) const;
		bool operator>=( ConstArrayType r ) const		{ return r <= *this; }
		bool operator<( ConstArrayType r ) const;
		bool operator>( ConstArrayType r ) const		{ return r < *this; }
		
		bool Touches( ConstArrayType r ) const;
		const ConstArrayType operator&( ConstArrayType r ) const;
		const ConstArrayType operator|( ConstArrayType r ) const;
		void operator&=( ConstArrayType r )				{ *this = *this & r; }
		void operator|=( ConstArrayType r )				{ *this = *this | r; }
  };

template < class Element >
inline void ConstArrayOf<Element>::Shorten( uint32 amount )
  {
	Assert( !Null() );
	Assert( amount <= length );
	start += amount;
	length -= amount;
  }

template < class Element >
inline const ConstArrayOf<Element> ConstArrayOf<Element>::Head( uint32 size ) const
  {
	Assert( !Null() );
	Assert( size <= length );
	return ConstArrayType( start, size );
  }

template < class Element >
inline const ConstArrayOf<Element> ConstArrayOf<Element>::Tail( uint32 position ) const
  {
	Assert( !Null() );
	Assert( position <= length );
	return ConstArrayType( start + position, length - position );
  }

template < class Element >
inline const ConstArrayOf<Element> ConstArrayOf<Element>::Middle( URange32 range ) const
  {
	Assert( !Null() );
	Assert( range.End() <= length );
	return ConstArrayType( start + range.Start(), range.Length() );
  }

#ifndef ConstArrayOf_cp
#include "ConstArrayOf.cp"
#endif

#endif
