// ConstData.h

#ifndef ConstData_h
#define ConstData_h

#ifndef Integers_h
#include "Integers.h"
#endif
#ifndef DebugMessage_h
#include "DebugMessage.h"
#endif
#ifndef Assert_h
#include "Assert.h"
#endif
#ifndef Range_h
#include "Range.h"
#endif
#ifndef ConstArrayOf_h
#include "ConstArrayOf.h"
#endif

// The includes above have been flattened so that weak compilers
// can cope better.

class Data;

/*
   A ConstData is a range of bytes, starting at Start() and
   not including End().  You can't change them.
*/

class ConstData
  {
	private:
		const uint8 *start;
		const uint8 *end;
	
	public:
		ConstData()		: start(0), end(0)	{}

		ConstData( const void *theStart, const void *theEnd )
		  : start( static_cast<const uint8 *>( theStart ) ),
			 end( static_cast<const uint8 *>( theEnd ) )
		  {
			Assert( theEnd >= theStart );
			Assert( theStart != 0 || theEnd == 0 );
		  }
		
		ConstData( const void *theStart, uint32 theLength )
		  : start( static_cast<const uint8 *>( theStart ) ),
			 end( static_cast<const uint8 *>( theStart ) + theLength )
		  {
			Assert( theStart != 0 || theLength == 0 );
		  }

		ConstData( ConstArrayOf<uint8> array )
		  : start( array.Start() ),
			 end( array.End() )
		  {}
		
		const uint8 *Start() const										{ return start; }
		const uint8 *End() const										{ return end; }

		uint32 Length() const											{ return uint32( end - start ); }
				
		uint32 BoundedLength( uint32 bound ) const				{ return ( Length() <= bound ) ? Length() : bound; }

		bool IsEmpty() const												{ return start == end; }
		bool Null() const													{ return start == 0; }
		
		const uint8& operator[]( uint32 i ) const					{ Assert( i < Length() ); return start[i]; }
		
		const ConstData Head( uint32 position ) const			{ Assert( position <= Length() ); return ConstData( start, start+position ); }
		const ConstData Tail( uint32 position ) const			{ Assert( position <= Length() ); return ConstData( start+position, end ); }
		const ConstData Middle( uint32 s, uint32 e ) const		{ Assert( s <= e );  Assert( e <= Length() ); return ConstData( start+s, start+e ); }
		const ConstData Middle( URange32 r ) const				{ Assert( r.End() <= Length() );  return ConstData( start+r.Start(), start+r.End() ); }
		
		void Truncate( uint32 position )								{ Assert( position <= Length() ); end = start + position; }
		void LimitLength( uint32 limit )								{ if ( limit < Length() )  end = start + limit; }
		
		void Shorten( uint32 amount )									{ Assert( amount <= Length() ); start += amount; }
		void Lengthen( uint32 amount )								{ end += amount; }
		
		bool StartsWith( ConstData ) const;
		bool EndsWith( ConstData ) const;
		bool Contains( ConstData ) const;
		
		bool Contains( const void *p ) const				{ return start <= p && p < end; }
		
		const ConstData operator>>( Data ) const;			// returns the remainder, for chaining
  };

bool operator==( const ConstData& a, const ConstData& b );
bool operator<( const ConstData& a, const ConstData& b );
bool operator<=( const ConstData& a, const ConstData& b );
		
inline bool operator!=( const ConstData& a, const ConstData& b )		{ return !( a == b ); }
inline bool operator>( const ConstData& a, const ConstData& b )		{ return !( a <= b ); }
inline bool operator>=( const ConstData& a, const ConstData& b )		{ return !( a < b ); }

int32 Compare( ConstData a, ConstData b );
int32 CompareReversed( ConstData a, ConstData b );

#endif
