// StackAllocatorBase.cp

#ifndef StackAllocatorBase_h
#include "StackAllocatorBase.h"
#endif
#ifndef MinMax_h
#include "MinMax.h"
#endif
#ifndef Alignment_h
#include "Alignment.h"
#endif
#ifndef TuneStackAllocators_h
#include "TuneStackAllocators.h"
#endif

StackAllocatorBase::StackAllocatorBase( Data baseSpace )
  : deepest( baseSpace, 0 ),
	 topPiece( &deepest ),
	 freePiece( 0 ),
	 top( baseSpace.Start() )
  {
	Assert( baseSpace.Length() > 0 );
	Assert( BestAlignment::IsAligned( baseSpace.Start() ) );
  }

StackAllocatorBase::~StackAllocatorBase()
  {
	Assert( topPiece == &deepest );
	Assert( top == topPiece->storage.Start() );
	ReleaseAll();
  }

void StackAllocatorBase::AllocatePiece( uint32 minimumSize )
  {
	Assert( !tuneStackAllocators );
	
	Assert( topPiece != 0 );
	Assert( freePiece == 0 );

	uint32 headSize = BestAlignment::Pad( sizeof(Piece) );
	uint32 tailSize = Max( 2 * topPiece->storage.Length(), minimumSize );
		
	void *pieceStorage = operator new( headSize + tailSize );
		
	Data storage( static_cast<uint8 *>( pieceStorage ) + headSize, tailSize );
		
	freePiece = new ( pieceStorage ) Piece( storage, topPiece );
  }

void StackAllocatorBase::ReleasePiece()
  {
	Assert( freePiece != 0 );
	Assert( freePiece != &deepest );
	freePiece->~Piece();
	operator delete( freePiece );
	freePiece = 0;
  }

void StackAllocatorBase::AddPiece( uint32 minimumSize )
  {
	Assert( topPiece != 0 );
	
	if ( freePiece != 0 && freePiece->storage.Length() < minimumSize )
		ReleasePiece();
			
	if ( freePiece == 0 )
		AllocatePiece( minimumSize );

	Assert( freePiece != 0 );
	topPiece = freePiece;
	freePiece = 0;
	top = topPiece->storage.Start();
  }

void StackAllocatorBase::RemovePiece()
  {
	Assert( topPiece != 0 );
	
	top = topPiece->storage.Start();
	
	if ( freePiece != 0 )
		ReleasePiece();
	
	Assert( freePiece == 0 );
	
	if ( topPiece->deeper != 0 )
	  {
		freePiece = topPiece;
		topPiece = topPiece->deeper;
		top = topPiece->storage.End();
	  }
  }

void *StackAllocatorBase::Allocate( size_t size )
  {
	Assert( size > 0 );
	if ( size <= 0 )
		size = 1;
	
	size += BestAlignment::Padding( size );
	
	if ( topPiece->storage.End() - top < size )
		AddPiece( size );
	
	Assert( topPiece->storage.End() - top >= size );
	
	void *result = top;
	top += size;
	return result;
  }

void StackAllocatorBase::Release( const void *p )
  {
	uint8 *toRelease = static_cast<uint8 *>( const_cast<void *>(p) );
	
	Assert( topPiece->storage.Contains( toRelease ) );
	Assert( toRelease < top );
	Assert( BestAlignment::IsAligned( toRelease ) );
	
	top = toRelease;
	if ( top == topPiece->storage.Start() )
		RemovePiece();
  }

void StackAllocatorBase::ReleaseTo( const void *p )
  {
	uint8 *toRelease = static_cast<uint8 *>( const_cast<void *>(p) );

	while (    toRelease < topPiece->storage.Start()
			  || toRelease > topPiece->storage.End() )
	  {
		Assert( topPiece != &deepest );
		RemovePiece();
	  }
	
	Assert( toRelease <= top );
	Assert( BestAlignment::IsAligned( toRelease ) );
	
	top = toRelease;
	if ( top == topPiece->storage.Start() )
		RemovePiece();
  }

void StackAllocatorBase::ReleaseAll()
  {
	while ( topPiece->deeper != 0 )
		RemovePiece();
	RemovePiece();
  }
