// StackBase.cp

#ifndef StackBase_h
#include "StackBase.h"
#endif
#ifndef Alignment_h
#include "Alignment.h"
#endif
#ifndef TuneStackAllocators_h
#include "TuneStackAllocators.h"
#endif

StackBase::StackBase( Data baseSpace, uint32 unit )
  : deepest( baseSpace, 0 ),
	 unitSize( unit ),
	 topPiece( &deepest ),
	 freePiece( 0 ),
	 depth( 0 ),
	 top( baseSpace.End() )
  {
	Assert( baseSpace.Length() > 0 );
	Assert( baseSpace.Length() % unit == 0 );
	Assert( BestAlignment::IsAligned( baseSpace.Start() ) );
	Assert( BestAlignment::IsAligned( baseSpace.End() ) );
  }

StackBase::~StackBase()
  {
	Assert( IsEmpty() );
	PopAll();
  }
		
uint8 *StackBase::AtDepth( uint32 d ) const
  {
	Assert( d < depth );
	Assert( topPiece != 0 );
	
	d *= unitSize;
	
	if ( d < topPiece->storage.End() - top )
		return top + d;
	
	d -= topPiece->storage.End() - top;
	
	for ( const Piece *piece = topPiece->deeper;
			piece != 0;
			piece = piece->deeper )
		if ( d < piece->storage.Length() )
			return piece->storage.Start() + d;
		 else
			d -= piece->storage.Length();
	
	Assert( 0 );
	return 0;
  }

void StackBase::AllocatePiece()
  {
	Assert( !tuneStackAllocators );
	
	Assert( topPiece != 0 );
	Assert( freePiece == 0 );

	uint32 headSize = BestAlignment::Pad( sizeof(Piece) );
	uint32 tailSize = 2 * topPiece->storage.Length();
		
	void *pieceStorage = operator new( headSize + tailSize );
		
	Data storage( static_cast<uint8 *>( pieceStorage ) + headSize, tailSize );
		
	freePiece = new ( pieceStorage ) Piece( storage, topPiece );
  }

void StackBase::ReleasePiece()
  {
	Assert( freePiece != 0 );
	Assert( freePiece != &deepest );
	freePiece->~Piece();
	operator delete( freePiece );
	freePiece = 0;
  }

void StackBase::AddPiece()
  {
	Assert( topPiece != 0 );
	Assert( top == topPiece->storage.Start() );
	
	if ( freePiece == 0 )
		AllocatePiece();

	Assert( freePiece != 0 );
	topPiece = freePiece;
	freePiece = 0;
	top = topPiece->storage.End();
  }

void StackBase::RemovePiece()
  {
	Assert( topPiece != 0 );
	Assert( top == topPiece->storage.End() );
	
	if ( freePiece != 0 )
		ReleasePiece();
	
	Assert( freePiece == 0 );
	
	if ( topPiece->deeper != 0 )
	  {
		freePiece = topPiece;
		topPiece = topPiece->deeper;
		top = topPiece->storage.Start();
	  }
  }

void *StackBase::Allocate()
  {
	Assert( topPiece != 0 );
	
	if ( top == topPiece->storage.Start() )
		AddPiece();
	
	Assert( topPiece != 0 );
	Assert( top > topPiece->storage.Start() );
	Assert( top - topPiece->storage.Start() >= unitSize );
	
	top -= unitSize;
	depth++;
	return top;
  }
		
void StackBase::Pop()
  {
	Assert( topPiece != 0 );
	Assert( depth > 0 );
	Assert( top < topPiece->storage.End() );
	Assert( topPiece->storage.End() - top >= unitSize );
	
	depth--;
	top += unitSize;
	
	if ( top == topPiece->storage.End() )
		RemovePiece();
  }

void StackBase::Pop( uint32 n )
  {
	Assert( n <= depth );
	
	if ( n == depth )
	  {
		PopAll();
		return;
	  }
	
	depth -= n;
	
	n *= unitSize;
	
	while ( n >= topPiece->storage.End() - top )
	  {
		n -= topPiece->storage.End() - top;
		top = topPiece->storage.End();
		RemovePiece();
	  }
	
	top += n;
  }

void StackBase::PopAll()
  {
	depth = 0;

	while ( top != topPiece->storage.End() )
	  {
		top = topPiece->storage.End();
		RemovePiece();
	  }
  }
